@@ -37,12 +37,15 @@ use crate::private;
 use crate::sql_type::OracleType;
 use crate::sql_type::ToSql;
 use crate::sql_value::BufferRowIndex;
+use crate::statement::QueryParams;
 use crate::to_odpi_str;
 use crate::to_rust_str;
 use crate::Connection;
 use crate::Error;
 use crate::Result;
 use crate::SqlValue;
+#[cfg(doc)]
+use crate::Statement;
 use crate::StatementType;
 use std::convert::TryFrom;
 use std::fmt;
@@ -50,9 +53,6 @@ use std::mem::MaybeUninit;
 use std::ptr;
 use std::slice;
 
-#[allow(unused_imports)]
-use crate::Statement; // for links in doc comments
-
 #[cfg(test)]
 const MINIMUM_TYPE_LENGTH: u32 = 1;
 #[cfg(not(test))]
@@ -119,6 +119,7 @@ pub struct BatchBuilder<'conn, 'sql> {
     batch_size: usize,
     with_batch_errors: bool,
     with_row_counts: bool,
+    query_params: QueryParams,
 }
 
 impl<'conn, 'sql> BatchBuilder<'conn, 'sql> {
@@ -133,6 +134,7 @@ impl<'conn, 'sql> BatchBuilder<'conn, 'sql> {
             batch_size: batch_size,
             with_batch_errors: false,
             with_row_counts: false,
+            query_params: QueryParams::new(),
         }
     }
 
@@ -155,9 +157,9 @@ impl<'conn, 'sql> BatchBuilder<'conn, 'sql> {
         let sql = to_odpi_str(self.sql);
         let mut handle: *mut dpiStmt = ptr::null_mut();
         chkerr!(
-            conn.ctxt,
+            conn.ctxt(),
             dpiConn_prepareStmt(
-                conn.handle.raw(),
+                conn.handle(),
                 0,
                 sql.ptr,
                 sql.len,
@@ -168,7 +170,7 @@ impl<'conn, 'sql> BatchBuilder<'conn, 'sql> {
         );
         let mut info = MaybeUninit::uninit();
         chkerr!(
-            conn.ctxt,
+            conn.ctxt(),
             dpiStmt_getInfo(handle, info.as_mut_ptr()),
             unsafe {
                 dpiStmt_release(handle);
@@ -186,9 +188,13 @@ impl<'conn, 'sql> BatchBuilder<'conn, 'sql> {
             return Err(Error::InvalidOperation(msg));
         };
         let mut num = 0;
-        chkerr!(conn.ctxt, dpiStmt_getBindCount(handle, &mut num), unsafe {
-            dpiStmt_release(handle);
-        });
+        chkerr!(
+            conn.ctxt(),
+            dpiStmt_getBindCount(handle, &mut num),
+            unsafe {
+                dpiStmt_release(handle);
+            }
+        );
         let bind_count = num as usize;
         let mut bind_names = Vec::with_capacity(bind_count);
         let mut bind_values = Vec::with_capacity(bind_count);
@@ -196,7 +202,7 @@ impl<'conn, 'sql> BatchBuilder<'conn, 'sql> {
             let mut names: Vec<*const i8> = vec![ptr::null_mut(); bind_count];
             let mut lengths = vec![0; bind_count];
             chkerr!(
-                conn.ctxt,
+                conn.ctxt(),
                 dpiStmt_getBindNames(handle, &mut num, names.as_mut_ptr(), lengths.as_mut_ptr()),
                 unsafe {
                     dpiStmt_release(handle);
@@ -205,7 +211,11 @@ impl<'conn, 'sql> BatchBuilder<'conn, 'sql> {
             bind_names = Vec::with_capacity(num as usize);
             for i in 0..(num as usize) {
                 bind_names.push(to_rust_str(names[i], lengths[i]));
-                bind_values.push(SqlValue::new(conn.ctxt));
+                bind_values.push(SqlValue::for_bind(
+                    conn.conn.clone(),
+                    self.query_params.clone(),
+                    batch_size,
+                ));
             }
         };
         Ok(Batch {
@@ -220,6 +230,7 @@ impl<'conn, 'sql> BatchBuilder<'conn, 'sql> {
             batch_size: batch_size,
             with_batch_errors: self.with_batch_errors,
             with_row_counts: self.with_row_counts,
+            query_params: self.query_params.clone(),
         })
     }
 }
@@ -234,7 +245,7 @@ impl<'conn, 'sql> BatchBuilder<'conn, 'sql> {
 ///
 /// # Usage
 ///
-/// 1. [`conn.batch(sql_stmt, batch_size).build()`][Connection#method.batch] to create [`Batch`].
+/// 1. [`conn.batch(sql_stmt, batch_size).build()`](Connection::batch) to create [`Batch`].
 /// 2. [`append_row()`](#method.append_row) for each row. Rows in the batch are sent to
 ///    the server when the number of appended rows reaches the batch size.  
 ///    **Note:** The "batch errors" option mentioned later changes this behavior.
@@ -306,7 +317,7 @@ impl<'conn, 'sql> BatchBuilder<'conn, 'sql> {
 ///
 /// **Note:** This feature is available only when both the client and the server are Oracle 12.1 or upper.
 ///
-/// [`BatchBuilder.with_batch_errors`][] changes
+/// [`BatchBuilder::with_batch_errors`] changes
 /// the behavior of `Batch` as follows:
 /// * `execute()` executes all rows in the batch and return an array of the error information
 ///   with row positions in the batch when the errors are caused by invalid data.
@@ -358,7 +369,7 @@ impl<'conn, 'sql> BatchBuilder<'conn, 'sql> {
 ///
 /// **Note:** This feature is available only when both the client and the server are Oracle 12.1 or upper.
 ///
-/// Use [`BatchBuilder.with_row_counts`][] and [`Batch.row_counts`][] to get affected rows
+/// Use [`BatchBuilder::with_row_counts`] and [`Batch::row_counts`] to get affected rows
 /// for each input row.
 ///
 /// ```
@@ -391,8 +402,8 @@ impl<'conn, 'sql> BatchBuilder<'conn, 'sql> {
 ///
 /// # Bind Parameter Types
 ///
-/// Parameter types are decided by the value of [`Batch.append_row`][], [`Batch.append_row_named`][]
-/// or [`Batch.set`][]; or by the type specified by [`Batch.set_type`][]. Once the
+/// Parameter types are decided by the value of [`Batch::append_row`], [`Batch::append_row_named`]
+/// or [`Batch::set`]; or by the type specified by [`Batch::set_type`]. Once the
 /// type is determined, there are no ways to change it except the following case.
 ///
 /// For user's convenience, when the length of character data types is too short,
@@ -414,15 +425,7 @@ impl<'conn, 'sql> BatchBuilder<'conn, 'sql> {
 /// ```
 /// Note that extending the internal buffer needs memory copy from existing buffer
 /// to newly allocated buffer. If you know the maximum data length, it is better
-/// to set the size by [`Batch.set_type`][].
-///
-/// [`BatchBuilder.with_batch_errors`]: BatchBuilder#method.with_batch_errors
-/// [`BatchBuilder.with_row_counts`]: BatchBuilder#method.with_row_counts
-/// [`Batch.append_row`]: Batch#method.append_row
-/// [`Batch.append_row_named`]: Batch#method.append_row_named
-/// [`Batch.row_counts`]: Batch#method.row_counts
-/// [`Batch.set`]: Batch#method.set
-/// [`Batch.set_type`]: Batch#method.set_type
+/// to set the size by [`Batch::set_type`].
 pub struct Batch<'conn> {
     pub(crate) conn: &'conn Connection,
     handle: *mut dpiStmt,
@@ -435,12 +438,13 @@ pub struct Batch<'conn> {
     batch_size: u32,
     with_batch_errors: bool,
     with_row_counts: bool,
+    query_params: QueryParams,
 }
 
 impl<'conn> Batch<'conn> {
     /// Closes the batch before the end of its lifetime.
     pub fn close(&mut self) -> Result<()> {
-        chkerr!(self.conn.ctxt, dpiStmt_close(self.handle, ptr::null(), 0));
+        chkerr!(self.conn.ctxt(), dpiStmt_close(self.handle, ptr::null(), 0));
         self.handle = ptr::null_mut();
         Ok(())
     }
@@ -493,7 +497,7 @@ impl<'conn> Batch<'conn> {
             return Ok(());
         }
         let mut exec_mode = DPI_MODE_EXEC_DEFAULT;
-        if self.conn.autocommit {
+        if self.conn.autocommit() {
             exec_mode |= DPI_MODE_EXEC_COMMIT_ON_SUCCESS;
         }
         if self.with_batch_errors {
@@ -503,19 +507,19 @@ impl<'conn> Batch<'conn> {
             exec_mode |= DPI_MODE_EXEC_ARRAY_DML_ROWCOUNTS;
         }
         chkerr!(
-            self.conn.ctxt,
+            self.conn.ctxt(),
             dpiStmt_executeMany(self.handle, exec_mode, self.batch_index)
         );
         if self.with_batch_errors {
             let mut errnum = 0;
             chkerr!(
-                self.conn.ctxt,
+                self.conn.ctxt(),
                 dpiStmt_getBatchErrorCount(self.handle, &mut errnum)
             );
             if errnum != 0 {
                 let mut errs = Vec::with_capacity(errnum as usize);
                 chkerr!(
-                    self.conn.ctxt,
+                    self.conn.ctxt(),
                     dpiStmt_getBatchErrors(self.handle, errnum, errs.as_mut_ptr())
                 );
                 unsafe { errs.set_len(errnum as usize) };
@@ -595,9 +599,9 @@ impl<'conn> Batch<'conn> {
                 bindidx
             )));
         }
-        self.bind_values[pos].init_handle(&self.conn.handle, oratype, self.batch_size)?;
+        self.bind_values[pos].init_handle(oratype)?;
         chkerr!(
-            self.conn.ctxt,
+            self.conn.ctxt(),
             bindidx.bind(self.handle, self.bind_values[pos].handle)
         );
         self.bind_types[pos] = Some(BindType::new(oratype));
@@ -645,13 +649,9 @@ impl<'conn> Batch<'conn> {
             // assume the type from the value
             let oratype = value.oratype(self.conn)?;
             let bind_type = BindType::new(&oratype);
-            self.bind_values[pos].init_handle(
-                &self.conn.handle,
-                bind_type.as_oratype().unwrap_or(&oratype),
-                self.batch_size,
-            )?;
+            self.bind_values[pos].init_handle(bind_type.as_oratype().unwrap_or(&oratype))?;
             chkerr!(
-                self.conn.ctxt,
+                self.conn.ctxt(),
                 bindidx.bind(self.handle, self.bind_values[pos].handle)
             );
             self.bind_types[pos] = Some(bind_type);
@@ -666,16 +666,16 @@ impl<'conn> Batch<'conn> {
                 let new_size = oratype_size(&new_oratype).ok_or(Error::DpiError(dberr))?;
                 bind_type.reset_size(new_size);
                 // allocate new bind handle.
-                let mut new_sql_value = SqlValue::new(self.conn.ctxt);
-                new_sql_value.init_handle(
-                    &self.conn.handle,
-                    bind_type.as_oratype().unwrap(),
+                let mut new_sql_value = SqlValue::for_bind(
+                    self.conn.conn.clone(),
+                    self.query_params.clone(),
                     self.batch_size,
-                )?;
+                );
+                new_sql_value.init_handle(bind_type.as_oratype().unwrap())?;
                 // copy values in old to new.
                 for idx in 0..self.batch_index {
                     chkerr!(
-                        self.conn.ctxt,
+                        self.conn.ctxt(),
                         dpiVar_copyData(
                             new_sql_value.handle,
                             idx,
@@ -687,7 +687,7 @@ impl<'conn> Batch<'conn> {
                 new_sql_value.buffer_row_index = BufferRowIndex::Owned(self.batch_index);
                 new_sql_value.set(value)?;
                 chkerr!(
-                    self.conn.ctxt,
+                    self.conn.ctxt(),
                     bindidx.bind(self.handle, new_sql_value.handle)
                 );
                 self.bind_values[pos] = new_sql_value;
@@ -711,7 +711,7 @@ impl<'conn> Batch<'conn> {
         let mut num_row_counts = 0;
         let mut row_counts = ptr::null_mut();
         chkerr!(
-            self.conn.ctxt,
+            self.conn.ctxt(),
             dpiStmt_getRowCounts(self.handle, &mut num_row_counts, &mut row_counts)
         );
         Ok(unsafe { slice::from_raw_parts(row_counts, num_row_counts as usize) }.to_vec())
