@@ -13,6 +13,7 @@
 // (ii) the Apache License v 2.0. (http://www.apache.org/licenses/LICENSE-2.0)
 //-----------------------------------------------------------------------------
 
+use std::borrow::ToOwned;
 use std::collections::HashMap;
 use std::fmt;
 use std::mem::{self, MaybeUninit};
@@ -23,17 +24,23 @@ use std::time::Duration;
 
 use crate::binding::*;
 use crate::chkerr;
-use crate::error::error_from_dpi_error;
 use crate::new_odpi_str;
+use crate::oci_attr::data_type::{AttrValue, DataType};
+use crate::oci_attr::handle::ConnHandle;
+use crate::oci_attr::handle::Server;
+use crate::oci_attr::mode::Read;
+use crate::oci_attr::mode::{ReadMode, WriteMode};
+use crate::oci_attr::OciAttr;
 use crate::sql_type::ObjectType;
 use crate::sql_type::ObjectTypeInternal;
 use crate::sql_type::ToSql;
 use crate::to_odpi_str;
-use crate::to_rust_slice;
 use crate::to_rust_str;
 use crate::util::duration_to_msecs;
 use crate::AssertSend;
 use crate::AssertSync;
+#[cfg(doc)]
+use crate::Batch;
 use crate::BatchBuilder;
 use crate::Context;
 use crate::DpiConn;
@@ -44,20 +51,24 @@ use crate::ResultSet;
 use crate::Row;
 use crate::RowValue;
 use crate::Statement;
+use crate::StatementBuilder;
 use crate::StmtParam;
 use crate::Version;
 
-#[allow(unused_imports)] // for links in doc comments
-use crate::Batch;
-
-const OCI_HTYPE_SERVER: u32 = 8;
+struct ServerStatus;
 const OCI_ATTR_SERVER_STATUS: u32 = 143;
 const OCI_SERVER_NOT_CONNECTED: u32 = 0;
 const OCI_SERVER_NORMAL: u32 = 1;
+unsafe impl OciAttr for ServerStatus {
+    type HandleType = Server;
+    type Mode = Read;
+    type DataType = u32;
+    const ATTR_NUM: u32 = OCI_ATTR_SERVER_STATUS;
+}
 
 /// Database startup mode
 ///
-/// See [`Connection.startup_database`](Connection#method.startup_database).
+/// See [`Connection::startup_database`]
 #[derive(Debug, Copy, Clone, PartialEq)]
 pub enum StartupMode {
     /// Shuts down a running instance (if there is any) using ABORT before
@@ -71,7 +82,7 @@ pub enum StartupMode {
 
 /// Database shutdown mode
 ///
-/// See [`Connection.shutdown_database`](Connection#method.shutdown_database).
+/// See [`Connection::shutdown_database`].
 #[derive(Debug, Copy, Clone, PartialEq)]
 pub enum ShutdownMode {
     /// Further connects are prohibited. Waits for users to disconnect from
@@ -100,13 +111,13 @@ pub enum ShutdownMode {
     Abort,
 
     /// Shuts down the database. Should be used only in the second call
-    /// to [`shutdown_database`](Connection#method.shutdown_database) after the database is closed and dismounted.
+    /// to [`Connection::shutdown_database`] after the database is closed and dismounted.
     Final,
 }
 
 /// [Administrative privilege](https://www.oracle.com/pls/topic/lookup?ctx=dblatest&id=GUID-633842B8-4B19-4F96-A757-783BF62825A7)
 ///
-/// See [Connector.privilege](struct.Connector.html#method.privilege).
+/// See [`Connector::privilege`].
 #[derive(Debug, Copy, Clone, PartialEq)]
 pub enum Privilege {
     /// Connects as [SYSDBA](https://www.oracle.com/pls/topic/lookup?ctx=dblatest&id=GUID-BD5D39D1-DBFF-400A-8645-355F8FB9CD31).
@@ -144,20 +155,18 @@ pub enum Purity {
 #[derive(Debug, Copy, Clone, PartialEq)]
 /// Connection status
 pub enum ConnStatus {
-    /// The connection is alive. See [`Connection.status`](Connection#method.status) for details.
+    /// The connection is alive. See [`Connection::status`] for details.
     Normal,
-    /// The connection has been terminated. See [`Connection.status`](Connection#method.status) for details.
+    /// The connection has been terminated. See [`Connection::status`] for details.
     NotConnected,
-    /// The connection has been closed by [`Connection.close`](Connection#method.close)
+    /// The connection has been closed by [`Connection::close`].
     Closed,
 }
 
 /// Builder data type to create Connection.
 ///
 /// When a connection can be established only with username, password
-/// and connect string, use [`Connection.connect`] instead.
-///
-/// [Connection.connect]: struct.Connection.html#method.connect
+/// and connect string, use [`Connection::connect`] instead.
 #[derive(Debug, Clone, PartialEq)]
 pub struct Connector {
     username: String,
@@ -241,7 +250,7 @@ impl Connector {
 
     /// Sets prelim auth mode to connect to an idle instance.
     ///
-    /// See [starting up a database](struct.Connection.html#method.startup_database).
+    /// See [starting up a database](Connection::startup_database).
     pub fn prelim_auth(&mut self, b: bool) -> &mut Connector {
         self.prelim_auth = b;
         self
@@ -461,26 +470,74 @@ impl Connector {
     }
 }
 
-/// Connection to an Oracle database
-pub struct Connection {
+pub(crate) type Conn = Arc<InnerConn>;
+
+pub(crate) struct InnerConn {
     pub(crate) ctxt: &'static Context,
     pub(crate) handle: DpiConn,
+    pub(crate) autocommit: Mutex<bool>,
+    pub(crate) objtype_cache: Mutex<HashMap<String, Arc<ObjectTypeInternal>>>,
     tag: String,
     tag_found: bool,
-    pub(crate) autocommit: bool,
-    pub(crate) objtype_cache: Mutex<HashMap<String, Arc<ObjectTypeInternal>>>,
 }
 
-impl AssertSync for Context {}
-impl AssertSend for Context {}
+impl InnerConn {
+    pub fn new(
+        ctxt: &'static Context,
+        handle: *mut dpiConn,
+        conn_params: &dpiConnCreateParams,
+    ) -> InnerConn {
+        InnerConn {
+            ctxt: ctxt,
+            handle: DpiConn::new(handle),
+            autocommit: Mutex::new(false),
+            objtype_cache: Mutex::new(HashMap::new()),
+            tag: to_rust_str(conn_params.outTag, conn_params.outTagLength),
+            tag_found: conn_params.outTagFound != 0,
+        }
+    }
+
+    pub fn autocommit(&self) -> bool {
+        *self.autocommit.lock().unwrap()
+    }
+
+    pub fn clear_object_type_cache(&self) -> Result<()> {
+        self.objtype_cache.lock()?.clear();
+        Ok(())
+    }
+}
+
+impl fmt::Debug for InnerConn {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        write!(
+            f,
+            "Conn {{ handle: {:?}, autocommit: {:?}",
+            self.handle.raw(),
+            self.autocommit,
+        )?;
+        if self.tag.len() != 0 {
+            write!(f, ", tag: {:?}", self.tag)?;
+        }
+        if self.tag_found {
+            write!(f, ", tag_found: {:?}", self.tag_found)?;
+        }
+        write!(f, " }}")
+    }
+}
+
+/// Connection to an Oracle database
+pub struct Connection {
+    pub(crate) conn: Conn,
+}
+
+impl AssertSync for Connection {}
+impl AssertSend for Connection {}
 
 impl Connection {
     /// Connects to an Oracle server using username, password and connect string.
     ///
     /// If you need to connect the server with additional parameters
-    /// such as SYSDBA privilege, use [Connector] instead.
-    ///
-    /// [Connector]: struct.Connector.html
+    /// such as SYSDBA privilege, use [`Connector`] instead.
     ///
     /// # Examples
     /// Connect to a local database.
@@ -544,15 +601,22 @@ impl Connection {
             )
         );
         Ok(Connection {
-            ctxt: ctxt,
-            handle: DpiConn::new(handle),
-            tag: to_rust_str(conn_params.outTag, conn_params.outTagLength),
-            tag_found: conn_params.outTagFound != 0,
-            autocommit: false,
-            objtype_cache: Mutex::new(HashMap::new()),
+            conn: Arc::new(InnerConn::new(ctxt, handle, &conn_params)),
         })
     }
 
+    pub(crate) fn from_conn(conn: Conn) -> Connection {
+        Connection { conn: conn }
+    }
+
+    pub(crate) fn ctxt(&self) -> &'static Context {
+        self.conn.ctxt
+    }
+
+    pub(crate) fn handle(&self) -> *mut dpiConn {
+        self.conn.handle.raw()
+    }
+
     /// Closes the connection before the end of lifetime.
     ///
     /// This fails when open statements or LOBs exist.
@@ -560,7 +624,91 @@ impl Connection {
         self.close_internal(DPI_MODE_CONN_CLOSE_DEFAULT, "")
     }
 
-    /// Prepares a statement
+    /// Creates [`StatementBuilder`][] to create a [`Statement`][]
+    ///
+    /// # Examples
+    ///
+    /// Executes a SQL statement with different parameters.
+    ///
+    /// ```no_run
+    /// # use oracle::*;
+    /// # let conn = Connection::connect("scott", "tiger", "")?;
+    /// let mut stmt = conn.statement("insert into emp(empno, ename) values (:id, :name)").build()?;
+    ///
+    /// let emp_list = [
+    ///     (7369, "Smith"),
+    ///     (7499, "Allen"),
+    ///     (7521, "Ward"),
+    /// ];
+    ///
+    /// // insert rows using positional parameters
+    /// for emp in &emp_list {
+    ///    stmt.execute(&[&emp.0, &emp.1])?;
+    /// }
+    ///
+    /// let emp_list = [
+    ///     (7566, "Jones"),
+    ///     (7654, "Martin"),
+    ///     (7698, "Blake"),
+    /// ];
+    ///
+    /// // insert rows using named parameters
+    /// for emp in &emp_list {
+    ///    stmt.execute_named(&[("id", &emp.0), ("name", &emp.1)])?;
+    /// }
+    /// # Ok::<(), Error>(())
+    /// ```
+    ///
+    /// Query methods in Connection allocate memory for 100 rows by default
+    /// to reduce the number of network round trips in case that many rows are
+    /// fetched. When 100 isn't preferable, use [`StatementBuilder::fetch_array_size`]
+    /// to customize it.
+    ///
+    /// ```no_run
+    /// # use oracle::*;
+    /// # let conn = Connection::connect("scott", "tiger", "")?;
+    /// // fetch top 10 rows.
+    /// let mut stmt = conn
+    ///     .statement("select empno, ename from emp order by empno fetch first 10 rows only")
+    ///     .fetch_array_size(10)
+    ///     .build()?;
+    /// for row_result in stmt.query_as::<(i32, String)>(&[])? {
+    ///     let (empno, ename) = row_result?;
+    ///     println!("empno: {}, ename: {}", empno, ename);
+    /// }
+    /// # Ok::<(), Error>(())
+    /// ```
+    ///
+    /// By default, a maximum of 2 rows are returned when the query is first
+    /// executed. To modify this, use [`StatementBuilder::prefetch_rows`] to customize
+    /// it. For more information on the difference between this and `fetch_array_size`,
+    /// see [this writeup](https://blog.dbi-services.com/arraysize-or-rowprefetch-in-sqlplus/)
+    /// or [this description](https://oracle.github.io/node-oracledb/doc/api.html#rowfetching).
+    ///
+    /// ```no_run
+    /// # use oracle::*;
+    /// # let conn = Connection::connect("scott", "tiger", "")?;
+    /// // fetch top 10 rows.
+    /// let mut stmt = conn
+    ///     .statement("select empno, ename from emp order by empno fetch first 10 rows only")
+    ///     .prefetch_rows(11)  // add one to avoid a round-trip to check for end-of-fetch
+    ///     .build()?;
+    /// for row_result in stmt.query_as::<(i32, String)>(&[])? {
+    ///     let (empno, ename) = row_result?;
+    ///     println!("empno: {}, ename: {}", empno, ename);
+    /// }
+    /// # Ok::<(), Error>(())
+    /// ```
+    pub fn statement<'conn, 'sql>(&'conn self, sql: &'sql str) -> StatementBuilder<'conn, 'sql> {
+        StatementBuilder::new(self, sql)
+    }
+
+    /// Creates a [`Statement`][] with parameters
+    ///
+    /// This will be marked as [deprecated][] at version 0.6.x. Please use [`statement(sql).build()`](#method.statement)
+    /// instead. The `statement()` method was added to follow [this API guideline](https://rust-lang.github.io/api-guidelines/type-safety.html#c-builder).
+    ///
+    /// [deprecated]: https://doc.rust-lang.org/edition-guide/rust-2018/the-compiler/an-attribute-for-deprecation.html
     ///
     /// # Examples
     ///
@@ -614,7 +762,7 @@ impl Connection {
     /// ```
     ///
     pub fn prepare(&self, sql: &str, params: &[StmtParam]) -> Result<Statement> {
-        Statement::new(self, sql, params)
+        Statement::from_params(self, sql, params)
     }
 
     /// Creates [BatchBuilder][]
@@ -634,12 +782,9 @@ impl Connection {
     ///
     /// [Query Methods]: https://github.com/kubo/rust-oracle/blob/master/docs/query-methods.md
     pub fn query(&self, sql: &str, params: &[&dyn ToSql]) -> Result<ResultSet<Row>> {
-        let mut rs = ResultSet::<Row>::from_conn(self, sql)?;
-        rs.stmt_boxed
-            .as_mut()
-            .unwrap()
-            .exec(params, true, "query")?;
-        Ok(rs)
+        let mut stmt = self.statement(sql).build()?;
+        stmt.exec(params, true, "query")?;
+        Ok(ResultSet::<Row>::from_stmt(stmt.stmt))
     }
 
     /// Executes a select statement using named parameters and returns a result set containing [`Row`]s.
@@ -648,12 +793,9 @@ impl Connection {
     ///
     /// [Query Methods]: https://github.com/kubo/rust-oracle/blob/master/docs/query-methods.md
     pub fn query_named(&self, sql: &str, params: &[(&str, &dyn ToSql)]) -> Result<ResultSet<Row>> {
-        let mut rs = ResultSet::<Row>::from_conn(self, sql)?;
-        rs.stmt_boxed
-            .as_mut()
-            .unwrap()
-            .exec_named(params, true, "query_named")?;
-        Ok(rs)
+        let mut stmt = self.statement(sql).build()?;
+        stmt.exec_named(params, true, "query_named")?;
+        Ok(ResultSet::<Row>::from_stmt(stmt.stmt))
     }
 
     /// Executes a select statement and returns a result set containing [`RowValue`]s.
@@ -665,12 +807,9 @@ impl Connection {
     where
         T: RowValue,
     {
-        let mut rs = ResultSet::from_conn(self, sql)?;
-        rs.stmt_boxed
-            .as_mut()
-            .unwrap()
-            .exec(params, true, "query_as")?;
-        Ok(rs)
+        let mut stmt = self.statement(sql).build()?;
+        stmt.exec(params, true, "query_as")?;
+        Ok(ResultSet::<T>::from_stmt(stmt.stmt))
     }
 
     /// Executes a select statement using named parameters and returns a result set containing [`RowValue`]s.
@@ -686,12 +825,9 @@ impl Connection {
     where
         T: RowValue,
     {
-        let mut rs = ResultSet::from_conn(self, sql)?;
-        rs.stmt_boxed
-            .as_mut()
-            .unwrap()
-            .exec_named(params, true, "query_as_named")?;
-        Ok(rs)
+        let mut stmt = self.statement(sql).build()?;
+        stmt.exec_named(params, true, "query_as_named")?;
+        Ok(ResultSet::<T>::from_stmt(stmt.stmt))
     }
 
     /// Gets one row from a query using positoinal bind parameters.
@@ -700,11 +836,11 @@ impl Connection {
     ///
     /// [Query Methods]: https://github.com/kubo/rust-oracle/blob/master/docs/query-methods.md
     pub fn query_row(&self, sql: &str, params: &[&dyn ToSql]) -> Result<Row> {
-        let mut stmt = self.prepare(sql, &[StmtParam::FetchArraySize(1)])?;
+        let mut stmt = self.statement(sql).fetch_array_size(1).build()?;
         if let Err(err) = stmt.query_row(params) {
             return Err(err);
         };
-        Ok(mem::replace(&mut stmt.row, None).unwrap())
+        Ok(mem::replace(&mut stmt.stmt.row, None).unwrap())
     }
 
     /// Gets one row from a query using named bind parameters.
@@ -713,11 +849,11 @@ impl Connection {
     ///
     /// [Query Methods]: https://github.com/kubo/rust-oracle/blob/master/docs/query-methods.md
     pub fn query_row_named(&self, sql: &str, params: &[(&str, &dyn ToSql)]) -> Result<Row> {
-        let mut stmt = self.prepare(sql, &[StmtParam::FetchArraySize(1)])?;
+        let mut stmt = self.statement(sql).fetch_array_size(1).build()?;
         if let Err(err) = stmt.query_row_named(params) {
             return Err(err);
         };
-        Ok(mem::replace(&mut stmt.row, None).unwrap())
+        Ok(mem::replace(&mut stmt.stmt.row, None).unwrap())
     }
 
     /// Gets one row from a query as specified type.
@@ -729,7 +865,7 @@ impl Connection {
     where
         T: RowValue,
     {
-        let mut stmt = self.prepare(sql, &[StmtParam::FetchArraySize(1)])?;
+        let mut stmt = self.statement(sql).fetch_array_size(1).build()?;
         stmt.query_row_as::<T>(params)
     }
 
@@ -742,11 +878,11 @@ impl Connection {
     where
         T: RowValue,
     {
-        let mut stmt = self.prepare(sql, &[StmtParam::FetchArraySize(1)])?;
+        let mut stmt = self.statement(sql).fetch_array_size(1).build()?;
         stmt.query_row_as_named::<T>(params)
     }
 
-    /// Prepares a statement, binds values by position and executes it in one call.
+    /// Creates a statement, binds values by position and executes it in one call.
     /// It will retunrs `Err` when the statemnet is a select statement.
     ///
     /// # Examples
@@ -764,12 +900,12 @@ impl Connection {
     /// # Ok::<(), Error>(())
     /// ```
     pub fn execute(&self, sql: &str, params: &[&dyn ToSql]) -> Result<Statement> {
-        let mut stmt = self.prepare(sql, &[])?;
+        let mut stmt = self.statement(sql).build()?;
         stmt.execute(params)?;
         Ok(stmt)
     }
 
-    /// Prepares a statement, binds values by name and executes it in one call.
+    /// Creates a statement, binds values by name and executes it in one call.
     /// It will retunrs `Err` when the statemnet is a select statement.
     ///
     /// The bind variable names are compared case-insensitively.
@@ -788,38 +924,38 @@ impl Connection {
     /// # Ok::<(), Error>(())
     /// ```
     pub fn execute_named(&self, sql: &str, params: &[(&str, &dyn ToSql)]) -> Result<Statement> {
-        let mut stmt = self.prepare(sql, &[])?;
+        let mut stmt = self.statement(sql).build()?;
         stmt.execute_named(params)?;
         Ok(stmt)
     }
 
     /// Commits the current active transaction
     pub fn commit(&self) -> Result<()> {
-        chkerr!(self.ctxt, dpiConn_commit(self.handle.raw()));
+        chkerr!(self.ctxt(), dpiConn_commit(self.handle()));
         Ok(())
     }
 
     /// Rolls back the current active transaction
     pub fn rollback(&self) -> Result<()> {
-        chkerr!(self.ctxt, dpiConn_rollback(self.handle.raw()));
+        chkerr!(self.ctxt(), dpiConn_rollback(self.handle()));
         Ok(())
     }
 
     /// Gets autocommit mode.
     /// It is false by default.
     pub fn autocommit(&self) -> bool {
-        self.autocommit
+        self.conn.autocommit()
     }
 
     /// Enables or disables autocommit mode.
     /// It is disabled by default.
     pub fn set_autocommit(&mut self, autocommit: bool) {
-        self.autocommit = autocommit;
+        *self.conn.autocommit.lock().unwrap() = autocommit;
     }
 
     /// Cancels execution of running statements in the connection
     pub fn break_execution(&self) -> Result<()> {
-        chkerr!(self.ctxt, dpiConn_breakExecution(self.handle.raw()));
+        chkerr!(self.ctxt(), dpiConn_breakExecution(self.handle()));
         Ok(())
     }
 
@@ -837,7 +973,7 @@ impl Connection {
     /// is executed, the cache clears.
     pub fn object_type(&self, name: &str) -> Result<ObjectType> {
         {
-            let guard = self.objtype_cache.lock()?;
+            let guard = self.conn.objtype_cache.lock()?;
             if let Some(rc_objtype) = guard.get(name) {
                 return Ok(ObjectType {
                     internal: rc_objtype.clone(),
@@ -847,12 +983,13 @@ impl Connection {
         let s = to_odpi_str(name);
         let mut handle = ptr::null_mut();
         chkerr!(
-            self.ctxt,
-            dpiConn_getObjectType(self.handle.raw(), s.ptr, s.len, &mut handle)
+            self.ctxt(),
+            dpiConn_getObjectType(self.handle(), s.ptr, s.len, &mut handle)
         );
-        let res = ObjectType::from_dpi_object_type(self.ctxt, DpiObjectType::new(handle));
+        let res = ObjectType::from_dpi_object_type(self.conn.clone(), DpiObjectType::new(handle));
         if let Ok(ref objtype) = res {
-            self.objtype_cache
+            self.conn
+                .objtype_cache
                 .lock()?
                 .insert(name.to_string(), objtype.internal.clone());
         };
@@ -863,13 +1000,12 @@ impl Connection {
     ///
     /// See also [`object_type`](#method.object_type).
     pub fn clear_object_type_cache(&self) -> Result<()> {
-        self.objtype_cache.lock()?.clear();
-        Ok(())
+        self.conn.clear_object_type_cache()
     }
 
     #[doc(hidden)]
     pub fn object_type_cache_len(&self) -> usize {
-        self.objtype_cache.lock().unwrap().len()
+        self.conn.objtype_cache.lock().unwrap().len()
     }
 
     /// Gets information about the server version
@@ -895,8 +1031,8 @@ impl Connection {
         let mut s = new_odpi_str();
         let mut ver = MaybeUninit::uninit();
         chkerr!(
-            self.ctxt,
-            dpiConn_getServerVersion(self.handle.raw(), &mut s.ptr, &mut s.len, ver.as_mut_ptr())
+            self.ctxt(),
+            dpiConn_getServerVersion(self.handle(), &mut s.ptr, &mut s.len, ver.as_mut_ptr())
         );
         Ok((
             Version::new_from_dpi_ver(unsafe { ver.assume_init() }),
@@ -915,9 +1051,9 @@ impl Connection {
         let old_password = to_odpi_str(old_password);
         let new_password = to_odpi_str(new_password);
         chkerr!(
-            self.ctxt,
+            self.ctxt(),
             dpiConn_changePassword(
-                self.handle.raw(),
+                self.handle(),
                 username.ptr,
                 username.len,
                 old_password.ptr,
@@ -934,16 +1070,16 @@ impl Connection {
     /// It checks the connection by making a network round-trip
     /// between the client and the server.
     ///
-    /// See also [Connection.status](struct.Connection.html#method.status).
+    /// See also [`Connection::status`].
     pub fn ping(&self) -> Result<()> {
-        chkerr!(self.ctxt, dpiConn_ping(self.handle.raw()));
+        chkerr!(self.ctxt(), dpiConn_ping(self.handle()));
         Ok(())
     }
 
     /// Gets the status of the connection.
     ///
     /// It returns `Ok(ConnStatus::Closed)` when the connection was closed
-    /// by [Connection.close](struct.Connection.html#method.close).
+    /// by [`Connection::close`].
     /// Otherwise see bellow.
     ///
     /// **Oracle client 12.2 and later:**
@@ -965,39 +1101,21 @@ impl Connection {
     /// `Ok(ConnStatus::NotConnected)`. There is no guarantee that the
     /// next network round-trip will go through.
     ///
-    /// See also [Connection.ping](struct.Connection.html#method.ping).
+    /// See also [`Connection::ping`].
     pub fn status(&self) -> Result<ConnStatus> {
-        unsafe {
-            let mut buf = MaybeUninit::uninit();
-            let mut len = mem::size_of::<u32>() as u32;
-            if dpiConn_getOciAttr(
-                self.handle.raw(),
-                OCI_HTYPE_SERVER,
-                OCI_ATTR_SERVER_STATUS,
-                buf.as_mut_ptr(),
-                &mut len,
-            ) == 0
-            {
-                let status = buf.assume_init().asUint32;
-                match status {
-                    OCI_SERVER_NOT_CONNECTED => Ok(ConnStatus::NotConnected),
-                    OCI_SERVER_NORMAL => Ok(ConnStatus::Normal),
-                    _ => Err(Error::InternalError(format!(
-                        "Unexpected server status: {}",
-                        status
-                    ))),
-                }
-            } else {
-                let mut err = MaybeUninit::uninit();
-                dpiContext_getError(self.ctxt.context, err.as_mut_ptr());
-                let err = err.assume_init();
-                let message = to_rust_slice(err.message, err.messageLength);
-                if message == b"DPI-1010: not connected" {
-                    Ok(ConnStatus::Closed)
-                } else {
-                    Err(error_from_dpi_error(&err))
-                }
+        match self.oci_attr::<ServerStatus>() {
+            Ok(status) => match status {
+                OCI_SERVER_NOT_CONNECTED => Ok(ConnStatus::NotConnected),
+                OCI_SERVER_NORMAL => Ok(ConnStatus::Normal),
+                _ => Err(Error::InternalError(format!(
+                    "Unexpected server status: {}",
+                    status
+                ))),
+            },
+            Err(Error::DpiError(err)) if err.message() == "DPI-1010: not connected" => {
+                Ok(ConnStatus::Closed)
             }
+            Err(err) => Err(err),
         }
     }
 
@@ -1005,15 +1123,15 @@ impl Connection {
     pub fn stmt_cache_size(&self) -> Result<u32> {
         let mut size = 0u32;
         chkerr!(
-            self.ctxt,
-            dpiConn_getStmtCacheSize(self.handle.raw(), &mut size)
+            self.ctxt(),
+            dpiConn_getStmtCacheSize(self.handle(), &mut size)
         );
         Ok(size)
     }
 
     /// Sets the statement cache size
     pub fn set_stmt_cache_size(&self, size: u32) -> Result<()> {
-        chkerr!(self.ctxt, dpiConn_setStmtCacheSize(self.handle.raw(), size));
+        chkerr!(self.ctxt(), dpiConn_setStmtCacheSize(self.handle(), size));
         Ok(())
     }
 
@@ -1023,8 +1141,8 @@ impl Connection {
     pub fn call_timeout(&self) -> Result<Option<Duration>> {
         let mut value = 0;
         chkerr!(
-            self.ctxt,
-            dpiConn_getCallTimeout(self.handle.raw(), &mut value)
+            self.ctxt(),
+            dpiConn_getCallTimeout(self.handle(), &mut value)
         );
         if value != 0 {
             Ok(Some(Duration::from_millis(value.into())))
@@ -1077,9 +1195,9 @@ impl Connection {
                     dur
                 )));
             }
-            chkerr!(self.ctxt, dpiConn_setCallTimeout(self.handle.raw(), msecs));
+            chkerr!(self.ctxt(), dpiConn_setCallTimeout(self.handle(), msecs));
         } else {
-            chkerr!(self.ctxt, dpiConn_setCallTimeout(self.handle.raw(), 0));
+            chkerr!(self.ctxt(), dpiConn_setCallTimeout(self.handle(), 0));
         }
         Ok(())
     }
@@ -1088,8 +1206,8 @@ impl Connection {
     pub fn current_schema(&self) -> Result<String> {
         let mut s = new_odpi_str();
         chkerr!(
-            self.ctxt,
-            dpiConn_getCurrentSchema(self.handle.raw(), &mut s.ptr, &mut s.len)
+            self.ctxt(),
+            dpiConn_getCurrentSchema(self.handle(), &mut s.ptr, &mut s.len)
         );
         Ok(s.to_string())
     }
@@ -1098,8 +1216,8 @@ impl Connection {
     pub fn set_current_schema(&self, current_schema: &str) -> Result<()> {
         let s = to_odpi_str(current_schema);
         chkerr!(
-            self.ctxt,
-            dpiConn_setCurrentSchema(self.handle.raw(), s.ptr, s.len)
+            self.ctxt(),
+            dpiConn_setCurrentSchema(self.handle(), s.ptr, s.len)
         );
         Ok(())
     }
@@ -1108,8 +1226,8 @@ impl Connection {
     pub fn edition(&self) -> Result<String> {
         let mut s = new_odpi_str();
         chkerr!(
-            self.ctxt,
-            dpiConn_getEdition(self.handle.raw(), &mut s.ptr, &mut s.len)
+            self.ctxt(),
+            dpiConn_getEdition(self.handle(), &mut s.ptr, &mut s.len)
         );
         Ok(s.to_string())
     }
@@ -1118,8 +1236,8 @@ impl Connection {
     pub fn external_name(&self) -> Result<String> {
         let mut s = new_odpi_str();
         chkerr!(
-            self.ctxt,
-            dpiConn_getExternalName(self.handle.raw(), &mut s.ptr, &mut s.len)
+            self.ctxt(),
+            dpiConn_getExternalName(self.handle(), &mut s.ptr, &mut s.len)
         );
         Ok(s.to_string())
     }
@@ -1128,8 +1246,8 @@ impl Connection {
     pub fn set_external_name(&self, external_name: &str) -> Result<()> {
         let s = to_odpi_str(external_name);
         chkerr!(
-            self.ctxt,
-            dpiConn_setExternalName(self.handle.raw(), s.ptr, s.len)
+            self.ctxt(),
+            dpiConn_setExternalName(self.handle(), s.ptr, s.len)
         );
         Ok(())
     }
@@ -1138,8 +1256,8 @@ impl Connection {
     pub fn internal_name(&self) -> Result<String> {
         let mut s = new_odpi_str();
         chkerr!(
-            self.ctxt,
-            dpiConn_getInternalName(self.handle.raw(), &mut s.ptr, &mut s.len)
+            self.ctxt(),
+            dpiConn_getInternalName(self.handle(), &mut s.ptr, &mut s.len)
         );
         Ok(s.to_string())
     }
@@ -1148,8 +1266,8 @@ impl Connection {
     pub fn set_internal_name(&self, internal_name: &str) -> Result<()> {
         let s = to_odpi_str(internal_name);
         chkerr!(
-            self.ctxt,
-            dpiConn_setInternalName(self.handle.raw(), s.ptr, s.len)
+            self.ctxt(),
+            dpiConn_setInternalName(self.handle(), s.ptr, s.len)
         );
         Ok(())
     }
@@ -1163,10 +1281,7 @@ impl Connection {
     /// [DBMS_APPLICATION_INFO.SET_MODULE]: https://www.oracle.com/pls/topic/lookup?ctx=dblatest&id=GUID-B2E2BD20-D91D-40DB-A3F6-37A853384F30
     pub fn set_module(&self, module: &str) -> Result<()> {
         let s = to_odpi_str(module);
-        chkerr!(
-            self.ctxt,
-            dpiConn_setModule(self.handle.raw(), s.ptr, s.len)
-        );
+        chkerr!(self.ctxt(), dpiConn_setModule(self.handle(), s.ptr, s.len));
         Ok(())
     }
 
@@ -1179,10 +1294,7 @@ impl Connection {
     /// [DBMS_APPLICATION_INFO.SET_ACTION]: https://www.oracle.com/pls/topic/lookup?ctx=dblatest&id=GUID-90DA860F-BFBE-4539-BA00-2279B02B8F26
     pub fn set_action(&self, action: &str) -> Result<()> {
         let s = to_odpi_str(action);
-        chkerr!(
-            self.ctxt,
-            dpiConn_setAction(self.handle.raw(), s.ptr, s.len)
-        );
+        chkerr!(self.ctxt(), dpiConn_setAction(self.handle(), s.ptr, s.len));
         Ok(())
     }
 
@@ -1196,8 +1308,8 @@ impl Connection {
     pub fn set_client_info(&self, client_info: &str) -> Result<()> {
         let s = to_odpi_str(client_info);
         chkerr!(
-            self.ctxt,
-            dpiConn_setClientInfo(self.handle.raw(), s.ptr, s.len)
+            self.ctxt(),
+            dpiConn_setClientInfo(self.handle(), s.ptr, s.len)
         );
         Ok(())
     }
@@ -1212,8 +1324,8 @@ impl Connection {
     pub fn set_client_identifier(&self, client_identifier: &str) -> Result<()> {
         let s = to_odpi_str(client_identifier);
         chkerr!(
-            self.ctxt,
-            dpiConn_setClientIdentifier(self.handle.raw(), s.ptr, s.len)
+            self.ctxt(),
+            dpiConn_setClientIdentifier(self.handle(), s.ptr, s.len)
         );
         Ok(())
     }
@@ -1232,7 +1344,7 @@ impl Connection {
     /// [Monitoring Database Operations]: https://www.oracle.com/pls/topic/lookup?ctx=dblatest&id=GUID-C941CE9D-97E1-42F8-91ED-4949B2B710BF
     pub fn set_db_op(&self, db_op: &str) -> Result<()> {
         let s = to_odpi_str(db_op);
-        chkerr!(self.ctxt, dpiConn_setDbOp(self.handle.raw(), s.ptr, s.len));
+        chkerr!(self.ctxt(), dpiConn_setDbOp(self.handle(), s.ptr, s.len));
         Ok(())
     }
 
@@ -1297,8 +1409,8 @@ impl Connection {
             };
         }
         chkerr!(
-            self.ctxt,
-            dpiConn_startupDatabase(self.handle.raw(), mode_num)
+            self.ctxt(),
+            dpiConn_startupDatabase(self.handle(), mode_num)
         );
         Ok(())
     }
@@ -1360,39 +1472,57 @@ impl Connection {
             ShutdownMode::Abort => DPI_MODE_SHUTDOWN_ABORT,
             ShutdownMode::Final => DPI_MODE_SHUTDOWN_FINAL,
         };
-        chkerr!(self.ctxt, dpiConn_shutdownDatabase(self.handle.raw(), mode));
+        chkerr!(self.ctxt(), dpiConn_shutdownDatabase(self.handle(), mode));
         Ok(())
     }
 
     #[doc(hidden)] // hiden until connection pooling is supported.
     pub fn tag(&self) -> &str {
-        &self.tag
+        &self.conn.tag
     }
 
     #[doc(hidden)] // hiden until connection pooling is supported.
     pub fn tag_found(&self) -> bool {
-        self.tag_found
+        self.conn.tag_found
     }
 
     fn close_internal(&self, mode: dpiConnCloseMode, tag: &str) -> Result<()> {
         let tag = to_odpi_str(tag);
         chkerr!(
-            self.ctxt,
-            dpiConn_close(self.handle.raw(), mode, tag.ptr, tag.len)
+            self.ctxt(),
+            dpiConn_close(self.handle(), mode, tag.ptr, tag.len)
         );
         Ok(())
     }
+
+    /// Gets an OCI handle attribute corresponding to the specified type parameter
+    /// See the [`oci_attr` module][crate::oci_attr] for details.
+    pub fn oci_attr<T>(&self) -> Result<<<T::DataType as DataType>::Type as ToOwned>::Owned>
+    where
+        T: OciAttr,
+        T::HandleType: ConnHandle,
+        T::Mode: ReadMode,
+    {
+        let attr_value = AttrValue::from_conn(self, <T::HandleType>::HANDLE_TYPE, <T>::ATTR_NUM);
+        unsafe { <T::DataType>::get(attr_value) }
+    }
+
+    /// Sets an OCI handle attribute corresponding to the specified type parameter
+    /// See the [`oci_attr` module][crate::oci_attr] for details.
+    pub fn set_oci_attr<T>(&mut self, value: &<T::DataType as DataType>::Type) -> Result<()>
+    where
+        T: OciAttr,
+        T::HandleType: ConnHandle,
+        T::Mode: WriteMode,
+    {
+        let mut attr_value =
+            AttrValue::from_conn(self, <T::HandleType>::HANDLE_TYPE, <T>::ATTR_NUM);
+        unsafe { <T::DataType>::set(&mut attr_value, value) }
+    }
 }
 
 impl fmt::Debug for Connection {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        write!(f, "Connection {{ handle: {:?}", self.handle.raw())?;
-        if self.tag.len() != 0 {
-            write!(f, ", tag: {:?}", self.tag)?;
-        }
-        if self.tag_found {
-            write!(f, ", tag_found: {:?}", self.tag_found)?;
-        }
-        write!(f, ", autocommit: {:?} }}", self.autocommit)
+        write!(f, "Connection {{ conn: {:?}", self.conn)
     }
 }
