@@ -21,17 +21,26 @@ use std::rc::Rc;
 use std::str;
 
 use crate::binding::*;
+use crate::binding_impl::DPI_MAX_BASIC_BUFFER_SIZE;
 use crate::chkerr;
+use crate::connection::Conn;
+use crate::sql_type::Bfile;
+use crate::sql_type::Blob;
+use crate::sql_type::Clob;
 use crate::sql_type::Collection;
 use crate::sql_type::FromSql;
 use crate::sql_type::IntervalDS;
 use crate::sql_type::IntervalYM;
 use crate::sql_type::NativeType;
+use crate::sql_type::Nclob;
 use crate::sql_type::Object;
 use crate::sql_type::ObjectType;
 use crate::sql_type::OracleType;
+use crate::sql_type::RefCursor;
 use crate::sql_type::Timestamp;
 use crate::sql_type::ToSql;
+use crate::statement::LobBindType;
+use crate::statement::QueryParams;
 use crate::to_rust_slice;
 use crate::to_rust_str;
 use crate::util::check_number_format;
@@ -39,7 +48,6 @@ use crate::util::parse_str_into_raw;
 use crate::util::set_hex_string;
 use crate::Connection;
 use crate::Context;
-use crate::DpiConn;
 use crate::Error;
 use crate::Result;
 
@@ -120,9 +128,9 @@ pub enum BufferRowIndex {
 /// determined by the column type.
 ///
 /// When this is a bind value in a SQL statement, the Oracle type is determined
-/// by [`ToSql.oratype`][ToSql#method.oratype].
+/// by [`ToSql::oratype`].
 pub struct SqlValue {
-    ctxt: &'static Context,
+    conn: Conn,
     pub(crate) handle: *mut dpiVar,
     data: *mut dpiData,
     native_type: NativeType,
@@ -131,33 +139,49 @@ pub struct SqlValue {
     pub(crate) buffer_row_index: BufferRowIndex,
     keep_bytes: Vec<u8>,
     keep_dpiobj: *mut dpiObject,
+    pub(crate) lob_bind_type: LobBindType,
+    pub(crate) query_params: QueryParams,
 }
 
 impl SqlValue {
-    // for column and bind values
-    pub(crate) fn new(ctxt: &'static Context) -> SqlValue {
+    fn new(
+        conn: Conn,
+        lob_bind_type: LobBindType,
+        query_params: QueryParams,
+        array_size: u32,
+    ) -> SqlValue {
         SqlValue {
-            ctxt: ctxt,
+            conn: conn,
             handle: ptr::null_mut(),
             data: ptr::null_mut(),
             native_type: NativeType::Int64,
             oratype: None,
-            array_size: 0,
+            array_size: array_size,
             buffer_row_index: BufferRowIndex::Owned(0),
             keep_bytes: Vec::new(),
             keep_dpiobj: ptr::null_mut(),
+            lob_bind_type: lob_bind_type,
+            query_params: query_params,
         }
     }
 
+    pub(crate) fn for_bind(conn: Conn, query_params: QueryParams, array_size: u32) -> SqlValue {
+        SqlValue::new(conn, LobBindType::Locator, query_params, array_size)
+    }
+
+    pub(crate) fn for_column(conn: Conn, query_params: QueryParams, array_size: u32) -> SqlValue {
+        SqlValue::new(conn, query_params.lob_bind_type, query_params, array_size)
+    }
+
     // for object type
     pub(crate) fn from_oratype(
-        ctxt: &'static Context,
+        conn: Conn,
         oratype: &OracleType,
         data: &mut dpiData,
     ) -> Result<SqlValue> {
         let (_, native_type, _, _) = oratype.var_create_param()?;
         Ok(SqlValue {
-            ctxt: ctxt,
+            conn: conn,
             handle: ptr::null_mut(),
             data: data as *mut dpiData,
             native_type: native_type,
@@ -166,16 +190,19 @@ impl SqlValue {
             buffer_row_index: BufferRowIndex::Owned(0),
             keep_bytes: Vec::new(),
             keep_dpiobj: ptr::null_mut(),
+            lob_bind_type: LobBindType::Locator,
+            query_params: QueryParams::new(),
         })
     }
 
-    fn handle_is_reusable(&self, oratype: &OracleType, array_size: u32) -> Result<bool> {
+    pub(crate) fn ctxt(&self) -> &'static Context {
+        self.conn.ctxt
+    }
+
+    fn handle_is_reusable(&self, oratype: &OracleType) -> Result<bool> {
         if self.handle.is_null() {
             return Ok(false);
         }
-        if self.array_size != array_size {
-            return Ok(false);
-        }
         let current_oratype = match self.oratype {
             Some(ref oratype) => oratype,
             None => return Ok(false),
@@ -197,13 +224,8 @@ impl SqlValue {
         }
     }
 
-    pub(crate) fn init_handle(
-        &mut self,
-        conn_handle: &DpiConn,
-        oratype: &OracleType,
-        array_size: u32,
-    ) -> Result<bool> {
-        if self.handle_is_reusable(oratype, array_size)? {
+    pub(crate) fn init_handle(&mut self, oratype: &OracleType) -> Result<bool> {
+        if self.handle_is_reusable(oratype)? {
             return Ok(false);
         }
         if !self.handle.is_null() {
@@ -212,16 +234,30 @@ impl SqlValue {
         self.handle = ptr::null_mut();
         let mut handle: *mut dpiVar = ptr::null_mut();
         let mut data: *mut dpiData = ptr::null_mut();
-        let (oratype_num, native_type, size, size_is_byte) = oratype.var_create_param()?;
+        let (oratype_num, native_type, size, size_is_byte) = match self.lob_bind_type {
+            LobBindType::Bytes => match oratype {
+                &OracleType::CLOB => &OracleType::Long,
+                &OracleType::NCLOB => {
+                    // When the size is larger than DPI_MAX_BASIC_BUFFER_SIZE, ODPI-C uses
+                    // a dynamic buffer instead of a fixed-size buffer.
+                    &OracleType::NVarchar2(DPI_MAX_BASIC_BUFFER_SIZE + 1)
+                }
+                &OracleType::BLOB => &OracleType::LongRaw,
+                &OracleType::BFILE => &OracleType::LongRaw,
+                _ => oratype,
+            },
+            LobBindType::Locator => oratype,
+        }
+        .var_create_param()?;
         let native_type_num = native_type.to_native_type_num();
         let object_type_handle = native_type.to_object_type_handle();
         chkerr!(
-            self.ctxt,
+            self.ctxt(),
             dpiConn_newVar(
-                conn_handle.raw(),
+                self.conn.handle.raw(),
                 oratype_num,
                 native_type_num,
-                array_size,
+                self.array_size,
                 size,
                 size_is_byte,
                 0,
@@ -234,7 +270,14 @@ impl SqlValue {
         self.data = data;
         self.native_type = native_type;
         self.oratype = Some(oratype.clone());
-        self.array_size = array_size;
+        if native_type_num == DPI_NATIVE_TYPE_STMT {
+            for i in 0..self.array_size {
+                let handle = unsafe { dpiData_getStmt(data.offset(i as isize)) };
+                if let Some(prefetch_rows) = self.query_params.prefetch_rows {
+                    chkerr!(self.ctxt(), dpiStmt_setPrefetchRows(handle, prefetch_rows));
+                }
+            }
+        }
         Ok(true)
     }
 
@@ -242,7 +285,7 @@ impl SqlValue {
         let mut num = 0;
         let mut data = ptr::null_mut();
         chkerr!(
-            self.ctxt,
+            self.ctxt(),
             dpiVar_getReturnedData(self.handle, 0, &mut num, &mut data)
         );
         if num != 0 {
@@ -277,10 +320,8 @@ impl SqlValue {
     ///    isn't `Option<FromSql>`, it returns `Err(Error::NullValue)`.
     /// 3. Converts the Oracle value to the rust value. The data type is converted
     ///    implicitly if required. For example string is converted to i64 by
-    ///    [`str.parse`][] if `get::<i64>()` is called for `VARCHAR2` columns.
+    ///    [`str::parse`] if `get::<i64>()` is called for `VARCHAR2` columns.
     ///    If the conversion fails, various errors are returned.
-    ///
-    /// [`str.parse`]: https://doc.rust-lang.org/std/primitive.str.html#method.parse
     pub fn get<T>(&self) -> Result<T>
     where
         T: FromSql,
@@ -323,6 +364,16 @@ impl SqlValue {
         }
     }
 
+    fn lob_locator_is_not_set<T>(&self, to_type: &str) -> Result<T> {
+        match self.oratype {
+            Some(_) => Err(Error::InvalidOperation(format!(
+                "Please use StatementBuilder.lob_locator() to fetch LOB data as {}",
+                to_type
+            ))),
+            None => Err(Error::UninitializedBindValue),
+        }
+    }
+
     fn check_not_null(&self) -> Result<()> {
         if self.is_null()? {
             Err(Error::NullValue)
@@ -477,7 +528,7 @@ impl SqlValue {
         while offset <= total_char_size {
             let mut read_len = bufsiz;
             chkerr!(
-                self.ctxt,
+                self.ctxt(),
                 dpiLob_readBytes(lob, offset, READ_CHAR_SIZE, bufptr, &mut read_len)
             );
             result.push_str(str::from_utf8(&buf[..(read_len as usize)])?);
@@ -496,7 +547,7 @@ impl SqlValue {
         let mut result: Vec<u8> = Vec::with_capacity(total_size as usize);
         let mut read_len = total_size;
         chkerr!(
-            self.ctxt,
+            self.ctxt(),
             dpiLob_readBytes(
                 lob,
                 1,
@@ -527,7 +578,7 @@ impl SqlValue {
         while offset <= total_size {
             let mut read_len = READ_SIZE;
             chkerr!(
-                self.ctxt,
+                self.ctxt(),
                 dpiLob_readBytes(lob, offset, READ_SIZE, bufptr, &mut read_len)
             );
             set_hex_string(&mut result, &buf[..(read_len as usize)]);
@@ -539,15 +590,15 @@ impl SqlValue {
     fn get_collection_unchecked(&self, objtype: &ObjectType) -> Result<Collection> {
         self.check_not_null()?;
         let dpiobj = unsafe { dpiData_getObject(self.data()) };
-        chkerr!(self.ctxt, dpiObject_addRef(dpiobj));
-        Ok(Collection::new(self.ctxt, dpiobj, objtype.clone()))
+        chkerr!(self.ctxt(), dpiObject_addRef(dpiobj));
+        Ok(Collection::new(self.conn.clone(), dpiobj, objtype.clone()))
     }
 
     fn get_object_unchecked(&self, objtype: &ObjectType) -> Result<Object> {
         self.check_not_null()?;
         let dpiobj = unsafe { dpiData_getObject(self.data()) };
-        chkerr!(self.ctxt, dpiObject_addRef(dpiobj));
-        Ok(Object::new(self.ctxt, dpiobj, objtype.clone()))
+        chkerr!(self.ctxt(), dpiObject_addRef(dpiobj));
+        Ok(Object::new(self.conn.clone(), dpiobj, objtype.clone()))
     }
 
     /// Gets the SQL value as bool. The native_type must be
@@ -562,12 +613,22 @@ impl SqlValue {
         let mut ptr = ptr::null();
         let mut len = 0;
         chkerr!(
-            self.ctxt,
+            self.ctxt(),
             dpiRowid_getStringValue((*self.data()).value.asRowid, &mut ptr, &mut len)
         );
         Ok(to_rust_str(ptr, len))
     }
 
+    fn get_lob_unchecked(&self) -> Result<*mut dpiLob> {
+        self.check_not_null()?;
+        unsafe { Ok(dpiData_getLOB(self.data())) }
+    }
+
+    fn get_stmt_unchecked(&self) -> Result<*mut dpiStmt> {
+        self.check_not_null()?;
+        unsafe { Ok(dpiData_getStmt(self.data())) }
+    }
+
     //
     // set_TYPE_unchecked methods
     //
@@ -613,7 +674,7 @@ impl SqlValue {
             }
         } else {
             chkerr!(
-                self.ctxt,
+                self.ctxt(),
                 dpiVar_setFromBytes(
                     self.handle,
                     self.buffer_row_index(),
@@ -685,8 +746,8 @@ impl SqlValue {
         let ptr = val.as_ptr() as *const i8;
         let len = val.len() as u64;
         let lob = unsafe { dpiData_getLOB(self.data()) };
-        chkerr!(self.ctxt, dpiLob_trim(lob, 0));
-        chkerr!(self.ctxt, dpiLob_writeBytes(lob, 1, ptr, len));
+        chkerr!(self.ctxt(), dpiLob_trim(lob, 0));
+        chkerr!(self.ctxt(), dpiLob_writeBytes(lob, 1, ptr, len));
         unsafe {
             (*self.data()).isNull = 0;
         }
@@ -697,8 +758,8 @@ impl SqlValue {
         let ptr = val.as_ptr() as *const i8;
         let len = val.len() as u64;
         let lob = unsafe { dpiData_getLOB(self.data()) };
-        chkerr!(self.ctxt, dpiLob_trim(lob, 0));
-        chkerr!(self.ctxt, dpiLob_writeBytes(lob, 1, ptr, len));
+        chkerr!(self.ctxt(), dpiLob_trim(lob, 0));
+        chkerr!(self.ctxt(), dpiLob_writeBytes(lob, 1, ptr, len));
         unsafe {
             (*self.data()).isNull = 0;
         }
@@ -717,7 +778,7 @@ impl SqlValue {
             self.keep_dpiobj = obj;
         } else {
             chkerr!(
-                self.ctxt,
+                self.ctxt(),
                 dpiVar_setFromObject(self.handle, self.buffer_row_index(), obj)
             );
         }
@@ -731,17 +792,30 @@ impl SqlValue {
         Ok(())
     }
 
-    /// Returns a duplicated value of self.
-    pub fn dup(&self, conn: &Connection) -> Result<SqlValue> {
-        self.dup_by_handle(&conn.handle)
+    fn set_lob_unchecked(&mut self, lob: *mut dpiLob) -> Result<()> {
+        chkerr!(
+            self.ctxt(),
+            dpiVar_setFromLob(self.handle, self.buffer_row_index(), lob)
+        );
+        Ok(())
     }
 
-    pub(crate) fn dup_by_handle(&self, conn_handle: &DpiConn) -> Result<SqlValue> {
-        let mut val = SqlValue::new(self.ctxt);
+    /// Returns a duplicated value of self.
+    pub fn dup(&self, _conn: &Connection) -> Result<SqlValue> {
+        self.dup_by_handle()
+    }
+
+    pub(crate) fn dup_by_handle(&self) -> Result<SqlValue> {
+        let mut val = SqlValue::new(
+            self.conn.clone(),
+            self.lob_bind_type,
+            self.query_params.clone(),
+            1,
+        );
         if let Some(ref oratype) = self.oratype {
-            val.init_handle(conn_handle, oratype, 1)?;
+            val.init_handle(oratype)?;
             chkerr!(
-                self.ctxt,
+                self.ctxt(),
                 dpiVar_copyData(val.handle, 0, self.handle, self.buffer_row_index()),
                 unsafe {
                     dpiVar_release(val.handle);
@@ -955,6 +1029,69 @@ impl SqlValue {
         }
     }
 
+    pub(crate) fn to_bfile(&self) -> Result<Bfile> {
+        if self.oratype == Some(OracleType::BFILE) {
+            match self.native_type {
+                NativeType::BLOB => {
+                    return Ok(Bfile::from_raw(self.ctxt(), self.get_lob_unchecked()?)?)
+                }
+                NativeType::Raw => return self.lob_locator_is_not_set("Bfile"),
+                _ => (),
+            }
+        }
+        self.invalid_conversion_to_rust_type("Bfile")
+    }
+
+    pub(crate) fn to_blob(&self) -> Result<Blob> {
+        if self.oratype == Some(OracleType::BLOB) {
+            match self.native_type {
+                NativeType::BLOB => {
+                    return Ok(Blob::from_raw(self.ctxt(), self.get_lob_unchecked()?)?)
+                }
+                NativeType::Raw => return self.lob_locator_is_not_set("Blob"),
+                _ => (),
+            }
+        }
+        self.invalid_conversion_to_rust_type("Blob")
+    }
+
+    pub(crate) fn to_clob(&self) -> Result<Clob> {
+        if self.oratype == Some(OracleType::CLOB) {
+            match self.native_type {
+                NativeType::CLOB => {
+                    return Ok(Clob::from_raw(self.ctxt(), self.get_lob_unchecked()?)?)
+                }
+                NativeType::Raw => return self.lob_locator_is_not_set("Clob"),
+                _ => (),
+            }
+        }
+        self.invalid_conversion_to_rust_type("Clob")
+    }
+
+    pub(crate) fn to_nclob(&self) -> Result<Nclob> {
+        if self.oratype == Some(OracleType::NCLOB) {
+            match self.native_type {
+                NativeType::CLOB => {
+                    return Ok(Nclob::from_raw(self.ctxt(), self.get_lob_unchecked()?)?)
+                }
+                NativeType::Raw => return self.lob_locator_is_not_set("Nclob"),
+                _ => (),
+            }
+        }
+        self.invalid_conversion_to_rust_type("Nclob")
+    }
+
+    pub(crate) fn to_ref_cursor(&self) -> Result<RefCursor> {
+        match self.native_type {
+            NativeType::Stmt => Ok(RefCursor::from_raw(
+                self.conn.clone(),
+                self.get_stmt_unchecked()?,
+                self.query_params.clone(),
+            )?),
+            _ => self.invalid_conversion_to_rust_type("RefCursor"),
+        }
+    }
+
     //
     // set_TYPE methods
     //
@@ -1091,11 +1228,55 @@ impl SqlValue {
         }
     }
 
+    pub(crate) fn set_bfile(&mut self, val: &Bfile) -> Result<()> {
+        if self.oratype == Some(OracleType::BFILE) {
+            match self.native_type {
+                NativeType::BLOB => return self.set_lob_unchecked(val.lob.handle),
+                NativeType::Raw => return self.lob_locator_is_not_set("Bfile"),
+                _ => (),
+            }
+        }
+        self.invalid_conversion_from_rust_type("Bfile")
+    }
+
+    pub(crate) fn set_blob(&mut self, val: &Blob) -> Result<()> {
+        if self.oratype == Some(OracleType::BLOB) {
+            match self.native_type {
+                NativeType::BLOB => return self.set_lob_unchecked(val.lob.handle),
+                NativeType::Raw => return self.lob_locator_is_not_set("Blob"),
+                _ => (),
+            }
+        }
+        self.invalid_conversion_from_rust_type("Blob")
+    }
+
+    pub(crate) fn set_clob(&mut self, val: &Clob) -> Result<()> {
+        if self.oratype == Some(OracleType::CLOB) {
+            match self.native_type {
+                NativeType::CLOB => return self.set_lob_unchecked(val.lob.handle),
+                NativeType::Raw => return self.lob_locator_is_not_set("Clob"),
+                _ => (),
+            }
+        }
+        self.invalid_conversion_from_rust_type("Clob")
+    }
+
+    pub(crate) fn set_nclob(&mut self, val: &Nclob) -> Result<()> {
+        if self.oratype == Some(OracleType::NCLOB) {
+            match self.native_type {
+                NativeType::CLOB => return self.set_lob_unchecked(val.lob.handle),
+                NativeType::Raw => return self.lob_locator_is_not_set("Nclob"),
+                _ => (),
+            }
+        }
+        self.invalid_conversion_from_rust_type("Nclob")
+    }
+
     /// The cloned value must not live longer than self.
     /// Otherwise it may cause access violation.
     pub(crate) fn unsafely_clone(&self) -> SqlValue {
         SqlValue {
-            ctxt: self.ctxt,
+            conn: self.conn.clone(),
             handle: ptr::null_mut(),
             data: self.data,
             native_type: self.native_type.clone(),
@@ -1104,6 +1285,8 @@ impl SqlValue {
             buffer_row_index: BufferRowIndex::Owned(0),
             keep_bytes: Vec::new(),
             keep_dpiobj: ptr::null_mut(),
+            lob_bind_type: self.lob_bind_type,
+            query_params: self.query_params.clone(),
         }
     }
 }
