@@ -13,23 +13,30 @@
 // (ii) the Apache License v 2.0. (http://www.apache.org/licenses/LICENSE-2.0)
 //-----------------------------------------------------------------------------
 
+use std::borrow::ToOwned;
 use std::cell::RefCell;
 use std::fmt;
-use std::mem::{self, MaybeUninit};
+use std::marker::PhantomData;
+use std::mem::MaybeUninit;
 use std::ptr;
 use std::rc::Rc;
 
 use crate::binding::*;
 use crate::chkerr;
-use crate::new_odpi_str;
+use crate::connection::Conn;
+use crate::oci_attr::data_type::{AttrValue, DataType};
+use crate::oci_attr::mode::{ReadMode, WriteMode};
+use crate::oci_attr::{self, OciAttr, SqlFnCode};
 use crate::private;
 use crate::sql_type::FromSql;
 use crate::sql_type::OracleType;
+use crate::sql_type::RefCursor;
 use crate::sql_type::ToSql;
 use crate::sql_value::BufferRowIndex;
 use crate::to_odpi_str;
 use crate::to_rust_str;
 use crate::Connection;
+use crate::Context;
 use crate::Error;
 use crate::Result;
 use crate::ResultSet;
@@ -37,14 +44,126 @@ use crate::Row;
 use crate::RowValue;
 use crate::SqlValue;
 
-const OCI_ATTR_SQLFNCODE: u32 = 10;
-
 // https://www.oracle.com/pls/topic/lookup?ctx=dblatest&id=GUID-A251CF91-EB9F-4DBC-8BB8-FB5EA92C20DE
 const SQLFNCODE_CREATE_TYPE: u16 = 77;
 const SQLFNCODE_ALTER_TYPE: u16 = 80;
 const SQLFNCODE_DROP_TYPE: u16 = 78;
 
-/// Parameters to prepare Statement.
+#[derive(Clone, Copy, Debug, PartialEq)]
+pub enum LobBindType {
+    Locator,
+    Bytes,
+}
+
+#[derive(Clone, Debug)]
+pub struct QueryParams {
+    pub fetch_array_size: u32,
+    pub prefetch_rows: Option<u32>,
+    pub lob_bind_type: LobBindType,
+}
+
+impl QueryParams {
+    pub fn new() -> QueryParams {
+        QueryParams {
+            fetch_array_size: DPI_DEFAULT_FETCH_ARRAY_SIZE,
+            prefetch_rows: None,
+            lob_bind_type: LobBindType::Bytes,
+        }
+    }
+}
+
+/// A builder to create a [`Statement`][] with various configuration
+pub struct StatementBuilder<'conn, 'sql> {
+    conn: &'conn Connection,
+    sql: &'sql str,
+    query_params: QueryParams,
+    scrollable: bool,
+    tag: String,
+}
+
+impl<'conn, 'sql> StatementBuilder<'conn, 'sql> {
+    pub(crate) fn new(conn: &'conn Connection, sql: &'sql str) -> StatementBuilder<'conn, 'sql> {
+        StatementBuilder {
+            conn: conn,
+            sql: sql,
+            query_params: QueryParams::new(),
+            scrollable: false,
+            tag: "".into(),
+        }
+    }
+
+    /// Changes the array size used for performing fetches.
+    ///
+    /// This specifies the number of rows allocated before performing
+    /// fetches. The default value is 100. Higher value reduces
+    /// the number of network round trips to fetch rows but requires
+    /// more memory. The preferable value depends on the query and
+    /// the environment.
+    ///
+    /// If the query returns only onw row, it is better to change
+    /// size to one.
+    ///
+    /// ```
+    /// # use oracle::Error;
+    /// # use oracle::test_util;
+    /// # let conn = test_util::connect()?;
+    /// let mut stmt = conn
+    ///     .statement("select StringCol from TestStrings where IntCol = :1")
+    ///     .fetch_array_size(1)
+    ///     .build()?;
+    /// assert_eq!(stmt.query_row_as::<String>(&[&1])?, "String 1");
+    /// assert_eq!(stmt.query_row_as::<String>(&[&2])?, "String 2");
+    /// # Ok::<(), Error>(())
+    /// ```
+    pub fn fetch_array_size<'a>(&'a mut self, size: u32) -> &'a mut StatementBuilder<'conn, 'sql> {
+        self.query_params.fetch_array_size = size;
+        self
+    }
+
+    /// The number of rows that will be prefetched by the Oracle Client
+    /// library when a query is executed. The default value is
+    /// DPI_DEFAULT_PREFETCH_ROWS (2). Increasing this value may reduce
+    /// the number of round-trips to the database that are required in
+    /// order to fetch rows, but at the cost of increasing memory
+    /// requirements.
+    /// Setting this value to 0 will disable prefetch completely,
+    /// which may be useful when the timing for fetching rows must be
+    /// controlled by the caller.
+    pub fn prefetch_rows<'a>(&'a mut self, size: u32) -> &'a mut StatementBuilder<'conn, 'sql> {
+        self.query_params.prefetch_rows = Some(size);
+        self
+    }
+
+    pub fn lob_locator<'a>(&'a mut self) -> &'a mut StatementBuilder<'conn, 'sql> {
+        self.query_params.lob_bind_type = LobBindType::Locator;
+        self
+    }
+
+    // make the visibility public when scrollable cursors is supported.
+    fn scrollable<'a>(&'a mut self, scrollable: bool) -> &'a mut StatementBuilder<'conn, 'sql> {
+        self.scrollable = scrollable;
+        self
+    }
+
+    // make the visibility public when statement caching is supported.
+    fn tag<'a, T>(&'a mut self, tag_name: T) -> &'a mut StatementBuilder<'conn, 'sql>
+    where
+        T: Into<String>,
+    {
+        self.tag = tag_name.into();
+        self
+    }
+
+    pub fn build(&self) -> Result<Statement<'conn>> {
+        Statement::new(self)
+    }
+}
+
+/// Parameters of [`Connection::prepare`]
+///
+/// No new variants are added to this enum in the future. That's because
+/// a new variant causes breaking changes. New configuration parameters
+/// are set via [`StatementBuilder`][] instead.
 #[derive(Debug, Clone, PartialEq)]
 pub enum StmtParam {
     /// The array size used for performing fetches.
@@ -66,7 +185,7 @@ pub enum StmtParam {
     Scrollable,
 }
 
-/// Statement type returned by [`Statement.statement_type`](Statement#method.statement_type).
+/// Statement type returned by [`Statement::statement_type`].
 #[derive(Debug, Copy, Clone, PartialEq)]
 pub enum StatementType {
     /// SELECT statement
@@ -165,50 +284,164 @@ impl fmt::Display for StatementType {
     }
 }
 
-/// Statement
-pub struct Statement<'conn> {
-    pub(crate) conn: &'conn Connection,
-    handle: *mut dpiStmt,
+#[derive(Debug)]
+pub(crate) struct Stmt {
+    pub(crate) conn: Conn,
+    pub(crate) handle: *mut dpiStmt,
     pub(crate) column_info: Vec<ColumnInfo>,
     pub(crate) row: Option<Row>,
     shared_buffer_row_index: Rc<RefCell<u32>>,
+    pub(crate) query_params: QueryParams,
+}
+
+impl Stmt {
+    pub(crate) fn new(conn: Conn, handle: *mut dpiStmt, query_params: QueryParams) -> Stmt {
+        Stmt {
+            conn: conn,
+            handle: handle,
+            column_info: Vec::new(),
+            row: None,
+            shared_buffer_row_index: Rc::new(RefCell::new(0)),
+            query_params: query_params,
+        }
+    }
+
+    pub(crate) fn ctxt(&self) -> &'static Context {
+        self.conn.ctxt
+    }
+
+    pub(crate) fn conn(&self) -> &Conn {
+        &self.conn
+    }
+
+    pub(crate) fn handle(&self) -> *mut dpiStmt {
+        self.handle
+    }
+
+    fn close(&mut self) -> Result<()> {
+        self.close_internal("")
+    }
+
+    fn close_internal(&mut self, tag: &str) -> Result<()> {
+        let tag = to_odpi_str(tag);
+
+        chkerr!(self.ctxt(), dpiStmt_close(self.handle, tag.ptr, tag.len));
+        self.handle = ptr::null_mut();
+        Ok(())
+    }
+
+    pub(crate) fn init_row(&mut self, num_cols: usize) -> Result<()> {
+        let mut column_names = Vec::with_capacity(num_cols);
+        let mut column_values = Vec::with_capacity(num_cols);
+        self.column_info = Vec::with_capacity(num_cols);
+
+        for i in 0..num_cols {
+            // set column info
+            let ci = ColumnInfo::new(self, i)?;
+            column_names.push(ci.name.clone());
+            self.column_info.push(ci);
+            // setup column value
+            let mut val = SqlValue::for_column(
+                self.conn.clone(),
+                self.query_params.clone(),
+                self.query_params.fetch_array_size,
+            );
+            val.buffer_row_index = BufferRowIndex::Shared(self.shared_buffer_row_index.clone());
+            let oratype = self.column_info[i].oracle_type();
+            let oratype_i64 = OracleType::Int64;
+            let oratype = match *oratype {
+                // When the column type is number whose prec is less than 18
+                // and the scale is zero, define it as int64.
+                OracleType::Number(prec, 0) if 0 < prec && prec < DPI_MAX_INT64_PRECISION as u8 => {
+                    &oratype_i64
+                }
+                _ => oratype,
+            };
+            val.init_handle(oratype)?;
+            chkerr!(
+                self.ctxt(),
+                dpiStmt_define(self.handle, (i + 1) as u32, val.handle)
+            );
+            column_values.push(val);
+        }
+        self.row = Some(Row::new(column_names, column_values)?);
+        Ok(())
+    }
+
+    pub fn next(&self) -> Option<Result<&Row>> {
+        let mut found = 0;
+        let mut buffer_row_index = 0;
+        if unsafe { dpiStmt_fetch(self.handle, &mut found, &mut buffer_row_index) } == 0 {
+            if found != 0 {
+                *self.shared_buffer_row_index.borrow_mut() = buffer_row_index;
+                // if self.row.is_none(), dpiStmt_fetch() returns non-zero.
+                Some(Ok(self.row.as_ref().unwrap()))
+            } else {
+                None
+            }
+        } else {
+            Some(Err(crate::error::error_from_context(self.ctxt())))
+        }
+    }
+
+    pub fn row_count(&self) -> Result<u64> {
+        let mut count = 0;
+        chkerr!(self.ctxt(), dpiStmt_getRowCount(self.handle, &mut count));
+        Ok(count)
+    }
+}
+
+impl Drop for Stmt {
+    fn drop(&mut self) {
+        unsafe { dpiStmt_release(self.handle) };
+    }
+}
+
+/// Statement
+#[derive(Debug)]
+pub struct Statement<'conn> {
+    pub(crate) stmt: Stmt,
     statement_type: StatementType,
     is_returning: bool,
     bind_count: usize,
     bind_names: Vec<String>,
     bind_values: Vec<SqlValue>,
-    fetch_array_size: u32,
+    phantom: PhantomData<&'conn ()>,
 }
 
 impl<'conn> Statement<'conn> {
-    pub(crate) fn new(
+    pub(crate) fn from_params(
         conn: &'conn Connection,
         sql: &str,
         params: &[StmtParam],
     ) -> Result<Statement<'conn>> {
-        let sql = to_odpi_str(sql);
-        let mut fetch_array_size = DPI_DEFAULT_FETCH_ARRAY_SIZE;
-        let mut scrollable = 0;
-        let mut tag = new_odpi_str();
+        let mut builder = conn.statement(sql);
         for param in params {
             match param {
                 &StmtParam::FetchArraySize(size) => {
-                    fetch_array_size = size;
+                    builder.fetch_array_size(size);
                 }
                 &StmtParam::Scrollable => {
-                    scrollable = 1;
+                    builder.scrollable(true);
                 }
                 &StmtParam::Tag(ref name) => {
-                    tag = to_odpi_str(name);
+                    builder.tag(name);
                 }
             }
         }
+        builder.build()
+    }
+
+    fn new<'sql>(builder: &StatementBuilder<'conn, 'sql>) -> Result<Statement<'conn>> {
+        let conn = builder.conn;
+        let sql = to_odpi_str(builder.sql);
+        let tag = to_odpi_str(&builder.tag);
         let mut handle: *mut dpiStmt = ptr::null_mut();
         chkerr!(
-            conn.ctxt,
+            conn.ctxt(),
             dpiConn_prepareStmt(
-                conn.handle.raw(),
-                scrollable,
+                conn.handle(),
+                if builder.scrollable { 1 } else { 0 },
                 sql.ptr,
                 sql.len,
                 tag.ptr,
@@ -218,7 +451,7 @@ impl<'conn> Statement<'conn> {
         );
         let mut info = MaybeUninit::uninit();
         chkerr!(
-            conn.ctxt,
+            conn.ctxt(),
             dpiStmt_getInfo(handle, info.as_mut_ptr()),
             unsafe {
                 dpiStmt_release(handle);
@@ -226,9 +459,13 @@ impl<'conn> Statement<'conn> {
         );
         let info = unsafe { info.assume_init() };
         let mut num = 0;
-        chkerr!(conn.ctxt, dpiStmt_getBindCount(handle, &mut num), unsafe {
-            dpiStmt_release(handle);
-        });
+        chkerr!(
+            conn.ctxt(),
+            dpiStmt_getBindCount(handle, &mut num),
+            unsafe {
+                dpiStmt_release(handle);
+            }
+        );
         let bind_count = num as usize;
         let mut bind_names = Vec::with_capacity(bind_count);
         let mut bind_values = Vec::with_capacity(bind_count);
@@ -236,7 +473,7 @@ impl<'conn> Statement<'conn> {
             let mut names: Vec<*const i8> = vec![ptr::null_mut(); bind_count];
             let mut lengths = vec![0; bind_count];
             chkerr!(
-                conn.ctxt,
+                conn.ctxt(),
                 dpiStmt_getBindNames(handle, &mut num, names.as_mut_ptr(), lengths.as_mut_ptr()),
                 unsafe {
                     dpiStmt_release(handle);
@@ -245,35 +482,39 @@ impl<'conn> Statement<'conn> {
             bind_names = Vec::with_capacity(num as usize);
             for i in 0..(num as usize) {
                 bind_names.push(to_rust_str(names[i], lengths[i]));
-                bind_values.push(SqlValue::new(conn.ctxt));
+                bind_values.push(SqlValue::for_bind(
+                    conn.conn.clone(),
+                    builder.query_params.clone(),
+                    1,
+                ));
             }
         };
         Ok(Statement {
-            conn: conn,
-            handle: handle,
-            column_info: Vec::new(),
-            row: None,
-            shared_buffer_row_index: Rc::new(RefCell::new(0)),
+            stmt: Stmt::new(conn.conn.clone(), handle, builder.query_params.clone()),
             statement_type: StatementType::from_enum(info.statementType),
             is_returning: info.isReturning != 0,
             bind_count: bind_count,
             bind_names: bind_names,
             bind_values: bind_values,
-            fetch_array_size: fetch_array_size,
+            phantom: PhantomData,
         })
     }
 
     /// Closes the statement before the end of lifetime.
     pub fn close(&mut self) -> Result<()> {
-        self.close_internal("")
+        self.stmt.close()
     }
 
-    fn close_internal(&mut self, tag: &str) -> Result<()> {
-        let tag = to_odpi_str(tag);
+    pub(crate) fn ctxt(&self) -> &'static Context {
+        self.conn().ctxt
+    }
 
-        chkerr!(self.conn.ctxt, dpiStmt_close(self.handle, tag.ptr, tag.len));
-        self.handle = ptr::null_mut();
-        Ok(())
+    pub(crate) fn conn(&self) -> &Conn {
+        &self.stmt.conn
+    }
+
+    pub(crate) fn handle(&self) -> *mut dpiStmt {
+        self.stmt.handle
     }
 
     /// Executes the prepared statement and returns a result set containing [`Row`]s.
@@ -283,7 +524,7 @@ impl<'conn> Statement<'conn> {
     /// [Query Methods]: https://github.com/kubo/rust-oracle/blob/master/docs/query-methods.md
     pub fn query(&mut self, params: &[&dyn ToSql]) -> Result<ResultSet<Row>> {
         self.exec(params, true, "query")?;
-        Ok(ResultSet::<Row>::new(self))
+        Ok(ResultSet::<Row>::new(&self.stmt))
     }
 
     /// Executes the prepared statement using named parameters and returns a result set containing [`Row`]s.
@@ -293,7 +534,7 @@ impl<'conn> Statement<'conn> {
     /// [Query Methods]: https://github.com/kubo/rust-oracle/blob/master/docs/query-methods.md
     pub fn query_named(&mut self, params: &[(&str, &dyn ToSql)]) -> Result<ResultSet<Row>> {
         self.exec_named(params, true, "query_named")?;
-        Ok(ResultSet::<Row>::new(self))
+        Ok(ResultSet::<Row>::new(&self.stmt))
     }
 
     /// Executes the prepared statement and returns a result set containing [`RowValue`]s.
@@ -306,7 +547,7 @@ impl<'conn> Statement<'conn> {
         T: RowValue,
     {
         self.exec(params, true, "query_as")?;
-        Ok(ResultSet::new(self))
+        Ok(ResultSet::new(&self.stmt))
     }
 
     /// Executes the prepared statement using named parameters and returns a result set containing [`RowValue`]s.
@@ -322,7 +563,7 @@ impl<'conn> Statement<'conn> {
         T: RowValue,
     {
         self.exec_named(params, true, "query_as_named")?;
-        Ok(ResultSet::new(self))
+        Ok(ResultSet::new(&self.stmt))
     }
 
     /// Gets one row from the prepared statement using positoinal bind parameters.
@@ -374,7 +615,7 @@ impl<'conn> Statement<'conn> {
     /// Binds values by position and executes the statement.
     /// It will retunrs `Err` when the statemnet is a select statement.
     ///
-    /// See also [`Connection.execute`](Connection#method.execute).
+    /// See also [`Connection::execute`].
     ///
     /// # Examples
     ///
@@ -400,7 +641,7 @@ impl<'conn> Statement<'conn> {
     /// Binds values by name and executes the statement.
     /// It will retunrs `Err` when the statemnet is a select statement.
     ///
-    /// See also [Connection.execute_named](Connection#method.execute_named).
+    /// See also [`Connection::execute_named`].
     ///
     /// # Examples
     ///
@@ -471,68 +712,35 @@ impl<'conn> Statement<'conn> {
     fn exec_common(&mut self) -> Result<()> {
         let mut num_query_columns = 0;
         let mut exec_mode = DPI_MODE_EXEC_DEFAULT;
-        if self.conn.autocommit {
+        if self.conn().autocommit() {
             exec_mode |= DPI_MODE_EXEC_COMMIT_ON_SUCCESS;
         }
         chkerr!(
-            self.conn.ctxt,
-            dpiStmt_setFetchArraySize(self.handle, self.fetch_array_size)
+            self.ctxt(),
+            dpiStmt_setFetchArraySize(self.handle(), self.stmt.query_params.fetch_array_size)
         );
+        if let Some(prefetch_rows) = self.stmt.query_params.prefetch_rows {
+            chkerr!(
+                self.ctxt(),
+                dpiStmt_setPrefetchRows(self.handle(), prefetch_rows)
+            );
+        }
         chkerr!(
-            self.conn.ctxt,
-            dpiStmt_execute(self.handle, exec_mode, &mut num_query_columns)
+            self.ctxt(),
+            dpiStmt_execute(self.handle(), exec_mode, &mut num_query_columns)
         );
         if self.is_ddl() {
-            let mut buf = MaybeUninit::uninit();
-            let mut len = mem::size_of::<u16>() as u32;
-            chkerr!(
-                self.conn.ctxt,
-                dpiStmt_getOciAttr(self.handle, OCI_ATTR_SQLFNCODE, buf.as_mut_ptr(), &mut len,)
-            );
-            let fncode = unsafe { buf.assume_init().asUint16 };
+            let fncode = self.oci_attr::<SqlFnCode>()?;
             match fncode {
                 SQLFNCODE_CREATE_TYPE | SQLFNCODE_ALTER_TYPE | SQLFNCODE_DROP_TYPE => {
-                    self.conn.clear_object_type_cache()?
+                    self.conn().clear_object_type_cache()?
                 }
                 _ => (),
             }
         }
         if self.statement_type == StatementType::Select {
-            if self.row.is_none() {
-                let num_cols = num_query_columns as usize;
-                let mut column_names = Vec::with_capacity(num_cols);
-                let mut column_values = Vec::with_capacity(num_cols);
-                self.column_info = Vec::with_capacity(num_cols);
-
-                for i in 0..num_cols {
-                    // set column info
-                    let ci = ColumnInfo::new(self, i)?;
-                    column_names.push(ci.name.clone());
-                    self.column_info.push(ci);
-                    // setup column value
-                    let mut val = SqlValue::new(self.conn.ctxt);
-                    val.buffer_row_index =
-                        BufferRowIndex::Shared(self.shared_buffer_row_index.clone());
-                    let oratype = self.column_info[i].oracle_type();
-                    let oratype_i64 = OracleType::Int64;
-                    let oratype = match *oratype {
-                        // When the column type is number whose prec is less than 18
-                        // and the scale is zero, define it as int64.
-                        OracleType::Number(prec, 0)
-                            if 0 < prec && prec < DPI_MAX_INT64_PRECISION as u8 =>
-                        {
-                            &oratype_i64
-                        }
-                        _ => oratype,
-                    };
-                    val.init_handle(&self.conn.handle, oratype, self.fetch_array_size)?;
-                    chkerr!(
-                        self.conn.ctxt,
-                        dpiStmt_define(self.handle, (i + 1) as u32, val.handle)
-                    );
-                    column_values.push(val);
-                }
-                self.row = Some(Row::new(self.conn, column_names, column_values)?);
+            if self.stmt.row.is_none() {
+                self.stmt.init_row(num_query_columns as usize)?;
             }
         }
         if self.is_returning {
@@ -617,10 +825,11 @@ impl<'conn> Statement<'conn> {
         I: BindIndex,
     {
         let pos = bindidx.idx(&self)?;
-        if self.bind_values[pos].init_handle(&self.conn.handle, &value.oratype(self.conn)?, 1)? {
+        let conn = Connection::from_conn(self.conn().clone());
+        if self.bind_values[pos].init_handle(&value.oratype(&conn)?)? {
             chkerr!(
-                self.conn.ctxt,
-                bindidx.bind(self.handle, self.bind_values[pos].handle)
+                self.ctxt(),
+                bindidx.bind(self.handle(), self.bind_values[pos].handle)
             );
         }
         self.bind_values[pos].set(value)
@@ -698,15 +907,13 @@ impl<'conn> Statement<'conn> {
     /// conn.execute("drop table people purge", &[])?;
     /// # Ok::<(), Error>(())
     /// ```
-    ///
-    /// [Statement.bind_value()]: #method.bind_value
     pub fn returned_values<I, T>(&self, bindidx: I) -> Result<Vec<T>>
     where
         I: BindIndex,
         T: FromSql,
     {
         let mut rows = 0;
-        chkerr!(self.conn.ctxt, dpiStmt_getRowCount(self.handle, &mut rows));
+        chkerr!(self.ctxt(), dpiStmt_getRowCount(self.handle(), &mut rows));
         if rows == 0 {
             return Ok(vec![]);
         }
@@ -722,28 +929,83 @@ impl<'conn> Statement<'conn> {
         Ok(vec)
     }
 
-    pub(crate) fn next(&self) -> Option<Result<&Row>> {
-        let mut found = 0;
-        let mut buffer_row_index = 0;
-        if unsafe { dpiStmt_fetch(self.handle, &mut found, &mut buffer_row_index) } == 0 {
-            if found != 0 {
-                *self.shared_buffer_row_index.borrow_mut() = buffer_row_index;
-                // if self.row.is_none(), dpiStmt_fetch() returns non-zero.
-                Some(Ok(self.row.as_ref().unwrap()))
-            } else {
-                None
-            }
-        } else {
-            Some(Err(crate::error::error_from_context(self.conn.ctxt)))
-        }
-    }
-
     /// Returns the number of rows fetched when the SQL statement is a query.
     /// Otherwise, the number of rows affected.
     pub fn row_count(&self) -> Result<u64> {
-        let mut count = 0;
-        chkerr!(self.conn.ctxt, dpiStmt_getRowCount(self.handle, &mut count));
-        Ok(count)
+        self.stmt.row_count()
+    }
+
+    /// Returns the next implicit result returned by [`dbms_sql.return_result()`]
+    /// in a PL/SQL block or a stored procedure.
+    ///
+    /// This feature is available when both the client and server are 12.1 or higher.
+    ///
+    /// [`dbms_sql.return_result()`]: https://www.oracle.com/pls/topic/lookup?ctx=dblatest&id=GUID-87562BF3-682C-48A7-B0C1-61075F19382A
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// # use oracle::Error;
+    /// # use oracle::test_util::{self, check_version, VER12_1};
+    /// # let conn = test_util::connect()?;
+    /// # if !check_version(&conn, &VER12_1, &VER12_1)? {
+    /// #     return Ok(()); // skip this test
+    /// # }
+    ///
+    /// let sql = r#"
+    /// declare
+    ///   cursor1 SYS_REFCURSOR;
+    ///   cursor2 SYS_REFCURSOR;
+    /// begin
+    ///   open cursor1 for select StringCol from TestStrings where IntCol = :1;
+    ///   -- return the first result set
+    ///   dbms_sql.return_result(cursor1);
+    ///
+    ///   open cursor2 for select StringCol from TestStrings where IntCol = :2;
+    ///   -- return the second result set
+    ///   dbms_sql.return_result(cursor2);
+    /// end;
+    /// "#;
+    ///
+    /// let mut stmt = conn.statement(sql).build()?;
+    /// stmt.execute(&[&1, &2])?;
+    ///
+    /// // Get the first result set.
+    /// let mut opt_cursor = stmt.implicit_result()?;
+    /// assert!(opt_cursor.is_some());
+    /// let mut cursor = opt_cursor.unwrap();
+    /// assert_eq!(cursor.query_row_as::<String>()?, "String 1");
+    ///
+    /// // Get the second result set.
+    /// let mut opt_cursor = stmt.implicit_result()?;
+    /// assert!(opt_cursor.is_some());
+    /// let mut cursor = opt_cursor.unwrap();
+    /// assert_eq!(cursor.query_row_as::<String>()?, "String 2");
+    ///
+    /// // No more result sets
+    /// let mut opt_cursor = stmt.implicit_result()?;
+    /// assert!(opt_cursor.is_none());
+    /// # Ok::<(), Error>(())
+    /// ```
+    pub fn implicit_result(&self) -> Result<Option<RefCursor>> {
+        let mut handle = ptr::null_mut();
+        chkerr!(
+            self.ctxt(),
+            dpiStmt_getImplicitResult(self.handle(), &mut handle)
+        );
+        if handle.is_null() {
+            Ok(None)
+        } else {
+            let cursor = RefCursor::from_raw(
+                self.stmt.conn.clone(),
+                handle,
+                self.stmt.query_params.clone(),
+            )?;
+            unsafe {
+                dpiStmt_release(handle);
+            }
+            Ok(Some(cursor))
+        }
     }
 
     /// Returns statement type
@@ -787,50 +1049,27 @@ impl<'conn> Statement<'conn> {
     pub fn is_returning(&self) -> bool {
         self.is_returning
     }
-}
 
-impl<'conn> Drop for Statement<'conn> {
-    fn drop(&mut self) {
-        unsafe { dpiStmt_release(self.handle) };
+    /// Gets an OCI handle attribute corresponding to the specified type parameter
+    /// See the [`oci_attr` module][crate::oci_attr] for details.
+    pub fn oci_attr<T>(&self) -> Result<<<T::DataType as DataType>::Type as ToOwned>::Owned>
+    where
+        T: OciAttr<HandleType = oci_attr::handle::Stmt>,
+        T::Mode: ReadMode,
+    {
+        let attr_value = AttrValue::from_stmt(self, <T>::ATTR_NUM);
+        unsafe { <T::DataType>::get(attr_value) }
     }
-}
 
-impl<'conn> fmt::Debug for Statement<'conn> {
-    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        write!(
-            f,
-            "Statement {{ handle: {:?}, conn: {:?}, stmt_type: {}",
-            self.handle,
-            self.conn,
-            self.statement_type()
-        )?;
-        if self.column_info.len() != 0 {
-            write!(f, ", colum_info: [")?;
-            for (idx, colinfo) in (&self.column_info).iter().enumerate() {
-                if idx != 0 {
-                    write!(f, ", ")?;
-                }
-                write!(
-                    f,
-                    "{} {}{}",
-                    colinfo.name(),
-                    colinfo.oracle_type(),
-                    if colinfo.nullable() { "" } else { " NOT NULL" }
-                )?;
-            }
-            write!(f, "], fetch_array_size: {}", self.fetch_array_size)?;
-        }
-        if self.bind_count != 0 {
-            write!(
-                f,
-                ", bind_count: {}, bind_names: {:?}, bind_values: {:?}",
-                self.bind_count, self.bind_names, self.bind_values
-            )?;
-        }
-        if self.is_returning {
-            write!(f, ", is_returning: true")?;
-        }
-        write!(f, " }}")
+    /// Sets an OCI handle attribute corresponding to the specified type parameter
+    /// See the [`oci_attr` module][crate::oci_attr] for details.
+    pub fn set_oci_attr<T>(&mut self, value: &<T::DataType as DataType>::Type) -> Result<()>
+    where
+        T: OciAttr<HandleType = oci_attr::handle::Stmt>,
+        T::Mode: WriteMode,
+    {
+        let mut attr_value = AttrValue::from_stmt(self, <T>::ATTR_NUM);
+        unsafe { <T::DataType>::set(&mut attr_value, value) }
     }
 }
 
@@ -878,16 +1117,16 @@ pub struct ColumnInfo {
 }
 
 impl ColumnInfo {
-    fn new(stmt: &Statement, idx: usize) -> Result<ColumnInfo> {
+    fn new(stmt: &Stmt, idx: usize) -> Result<ColumnInfo> {
         let mut info = MaybeUninit::uninit();
         chkerr!(
-            stmt.conn.ctxt,
-            dpiStmt_getQueryInfo(stmt.handle, (idx + 1) as u32, info.as_mut_ptr())
+            stmt.ctxt(),
+            dpiStmt_getQueryInfo(stmt.handle(), (idx + 1) as u32, info.as_mut_ptr())
         );
         let info = unsafe { info.assume_init() };
         Ok(ColumnInfo {
             name: to_rust_str(info.name, info.nameLength),
-            oracle_type: OracleType::from_type_info(stmt.conn.ctxt, &info.typeInfo)?,
+            oracle_type: OracleType::from_type_info(stmt.conn(), &info.typeInfo)?,
             nullable: info.nullOk != 0,
         })
     }
@@ -934,7 +1173,7 @@ pub trait BindIndex: private::Sealed {
 impl BindIndex for usize {
     fn idx(&self, stmt: &Statement) -> Result<usize> {
         let num = stmt.bind_count();
-        if 0 < num && *self <= num {
+        if 0 < num && 1 <= *self && *self <= num {
             Ok(*self - 1)
         } else {
             Err(Error::InvalidBindIndex(*self))
@@ -991,3 +1230,92 @@ impl<'a> ColumnIndex for &'a str {
         Err(Error::InvalidColumnName((*self).to_string()))
     }
 }
+
+#[cfg(test)]
+mod tests {
+    use super::*;
+    use crate::test_util;
+
+    #[test]
+    fn column_info() -> Result<()> {
+        let conn = test_util::connect()?;
+        let mut stmt = conn.statement("select * from TestDataTypes").build()?;
+        let rows = stmt.query(&[])?;
+        let colinfo = rows.column_info();
+        assert_eq!(colinfo[0].name(), "STRINGCOL");
+        assert_eq!(colinfo[0].oracle_type(), &OracleType::Varchar2(100));
+        assert_eq!(colinfo[1].name(), "UNICODECOL");
+        assert_eq!(colinfo[1].oracle_type(), &OracleType::NVarchar2(100));
+        assert_eq!(colinfo[2].name(), "FIXEDCHARCOL");
+        assert_eq!(colinfo[2].oracle_type(), &OracleType::Char(100));
+        assert_eq!(colinfo[3].name(), "FIXEDUNICODECOL");
+        assert_eq!(colinfo[3].oracle_type(), &OracleType::NChar(100));
+        assert_eq!(colinfo[4].name(), "RAWCOL");
+        assert_eq!(colinfo[4].oracle_type(), &OracleType::Raw(30));
+        assert_eq!(colinfo[5].name(), "FLOATCOL");
+        assert_eq!(colinfo[5].oracle_type(), &OracleType::Float(126));
+        assert_eq!(colinfo[6].name(), "DOUBLEPRECCOL");
+        assert_eq!(colinfo[6].oracle_type(), &OracleType::Float(126));
+        assert_eq!(colinfo[7].name(), "INTCOL");
+        assert_eq!(colinfo[7].oracle_type(), &OracleType::Number(9, 0));
+        assert_eq!(colinfo[8].name(), "NUMBERCOL");
+        assert_eq!(colinfo[8].oracle_type(), &OracleType::Number(9, 2));
+        assert_eq!(colinfo[9].name(), "DATECOL");
+        assert_eq!(colinfo[9].oracle_type(), &OracleType::Date);
+        assert_eq!(colinfo[10].name(), "TIMESTAMPCOL");
+        assert_eq!(colinfo[10].oracle_type(), &OracleType::Timestamp(6));
+        assert_eq!(colinfo[11].name(), "TIMESTAMPTZCOL");
+        assert_eq!(colinfo[11].oracle_type(), &OracleType::TimestampTZ(6));
+        assert_eq!(colinfo[12].name(), "TIMESTAMPLTZCOL");
+        assert_eq!(colinfo[12].oracle_type(), &OracleType::TimestampLTZ(6));
+        assert_eq!(colinfo[13].name(), "INTERVALDSCOL");
+        assert_eq!(colinfo[13].oracle_type(), &OracleType::IntervalDS(2, 6));
+        assert_eq!(colinfo[14].name(), "INTERVALYMCOL");
+        assert_eq!(colinfo[14].oracle_type(), &OracleType::IntervalYM(2));
+        assert_eq!(colinfo[15].name(), "BINARYFLTCOL");
+        assert_eq!(colinfo[15].oracle_type(), &OracleType::BinaryFloat);
+        assert_eq!(colinfo[16].name(), "BINARYDOUBLECOL");
+        assert_eq!(colinfo[16].oracle_type(), &OracleType::BinaryDouble);
+        assert_eq!(colinfo[17].name(), "CLOBCOL");
+        assert_eq!(colinfo[17].oracle_type(), &OracleType::CLOB);
+        assert_eq!(colinfo[18].name(), "NCLOBCOL");
+        assert_eq!(colinfo[18].oracle_type(), &OracleType::NCLOB);
+        assert_eq!(colinfo[19].name(), "BLOBCOL");
+        assert_eq!(colinfo[19].oracle_type(), &OracleType::BLOB);
+        assert_eq!(colinfo[20].name(), "BFILECOL");
+        assert_eq!(colinfo[20].oracle_type(), &OracleType::BFILE);
+        assert_eq!(colinfo[21].name(), "LONGCOL");
+        assert_eq!(colinfo[21].oracle_type(), &OracleType::Long);
+        assert_eq!(colinfo[22].name(), "UNCONSTRAINEDCOL");
+        assert_eq!(colinfo[22].oracle_type(), &OracleType::Number(0, -127));
+        assert_eq!(colinfo[23].name(), "SIGNEDINTCOL");
+        assert_eq!(colinfo[23].oracle_type(), &OracleType::Number(38, 0));
+        assert_eq!(colinfo[24].name(), "SUBOBJECTCOL");
+        assert_eq!(
+            colinfo[24].oracle_type().to_string(),
+            OracleType::Object(conn.object_type("UDT_SUBOBJECT")?).to_string()
+        );
+        assert_eq!(colinfo.len(), 25);
+
+        let mut stmt = conn.statement("select * from TestLongRaws").build()?;
+        let rows = stmt.query(&[])?;
+        let colinfo = rows.column_info();
+        assert_eq!(colinfo[0].name(), "INTCOL");
+        assert_eq!(colinfo[0].oracle_type(), &OracleType::Number(9, 0));
+        assert_eq!(colinfo[1].name(), "LONGRAWCOL");
+        assert_eq!(colinfo[1].oracle_type(), &OracleType::LongRaw);
+        assert_eq!(colinfo.len(), 2);
+
+        let mut stmt = conn.statement("select * from TestXml").build()?;
+        let rows = stmt.query(&[])?;
+        let colinfo = rows.column_info();
+        assert_eq!(colinfo[0].name(), "INTCOL");
+        assert_eq!(colinfo[0].oracle_type(), &OracleType::Number(9, 0));
+        assert_eq!(colinfo[1].name(), "XMLCOL");
+        // xmltype is reported as long. (https://github.com/oracle/odpi/commit/d51e9c496381427dd09f7fb0cbaab3e152667026)
+        assert_eq!(colinfo[1].oracle_type(), &OracleType::Long);
+        assert_eq!(colinfo.len(), 2);
+
+        Ok(())
+    }
+}
